/* hp2100_di.c: HP 12821A HP-IB Disc Interface simulator

   Copyright (c) 2010-2022, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from the author.

   DI           12821A Disc Interface

   27-Mar-22    JDB     Added "da_bus_initialize" to the bus handler table
   11-Mar-22    JDB     Moved execution initialization to "di_initializer"
   13-Jul-21    JDB     Added "di_bus_trace" global routine
   02-Jul-21    JDB     Added "tag_reg" pointer to update tag queue pointer
   25-Jun-21    JDB     Removed "di_poll_response"; now handled by devices
   16-Jun-21    JDB     Added bus INITIALIZER definition
                        Reassigned trace flags
   18-May-21    JDB     Logically separated SIMH unit numbers from bus addresses
   04-May-21    JDB     Removed dummy DC device
   09-Dec-20    JDB     Replaced "io_assert" with "io_assert_ENF/SIR"
   02-Dec-20    JDB     RESET no longer presets the interface
   11-Jul-18    JDB     Revised I/O model
   15-Mar-17    JDB     Converted debug fprintfs to tpprintfs
   10-Mar-17    JDB     Added IOBUS to the debug table
   17-Jan-17    JDB     Changed to use new byte accessors in hp2100_defs.h
   13-May-16    JDB     Modified for revised SCP API function parameter types
   24-Dec-14    JDB     Added casts for explicit downward conversions
                        Removed redundant global declarations
   13-Feb-12    JDB     First release
   15-Dec-11    JDB     Added dummy DC device for diagnostics
   09-Oct-10    JDB     Created DI simulation

   References:
     - HP 12821A Disc Interface Installation and Service Manual
         (12821-90006, February 1985)
     - IEEE Standard Digital Interface for Programmable Instrumentation
         (IEEE-488A-1980, September 1979)


   The 12821A Disc Interface is a high-speed interface to the Hewlett-Packard
   Interface Bus (HP-IB, formalized as IEEE Std. 488-1978).  It is used to
   interface HP-IB disc and tape devices, such as the HP 7906H, 7908A, and
   7974A, to an HP 1000 computer running RTE-IVB or RTE-6/VM.  Three device
   command protocols are supported by the RTE I/O drivers: Amigo discs by driver
   DVA32, CS/80 discs and cartridge tapes by DVM33, and Amigo reel-to-reel tapes
   by DVS23.  While RTE-6/VM supports all three drivers, only DVA32 for Amigo
   discs is available for RTE-IVB.

   This simulator is written to allow the definition of multiple DI cards in a
   system.  As RTE requires that all devices connected to a given interface card
   use the same driver, simultaneous use of Amigo and CS/80 devices requires
   separate interface cards.

   In an RTE environment, the 12821A is the system controller.  While
   electrically compatible with the HP-IB specification and capable of receiving
   addressing commands from the bus, the 12821A does not use the full IEEE-488
   protocol.  Card talker and listener states are set by bits in the control
   register, rather than by receiving talk and listen commands over the bus, and
   the RTE drivers do not address the controller when setting up transfers to or
   from the bus devices.  The bus address of the card can be set via DIP
   switches, but this feature is used only by the diagnostic.

   The card supports packed and unpacked transfers across the bus.  Up to four
   devices can be connected to each card; this limit is imposed by the maximum
   electrical loading on the bus compatible with the high data rate.  This
   restriction does not exist in the simulator, however.

   The 12821A has a 16-word FIFO buffer and can sustain DCPC transfers of one
   megabyte per second.  Burst transfers by the CPU to fill or empty the FIFO
   can run at the full bandwidth of the I/O backplane.  This can hold off
   lower-priority devices for 10-15 microseconds until the card slows down to
   the rate of the disc or tape.

   Card assembly 12821-60003 was revised to add a DCPC pacing option.  Placing
   jumper W1 in position A inhibits SRQ for one I/O cycle in six to allow a
   lower-priority interface card to transfer one word.  Position B allows SRQ to
   assert continuously as it does when using the earlier card assembly
   12821-60001.

   This simulation is logically partitioned into three sets of functions: the
   interface card simulation, the HP-IB bus simulation, and the device
   simulation.  This is the card simulation and the card portion of the HP-IB
   simulation.  Separate modules for the tape and disc devices contain the
   device simulations and the device portions of the HP-IB simulations.  The
   division is conceptually at the HP-IB cable that connects the interface to
   the devices.  A detailed description of the logical interface is presented in
   the comments preceding the global bus routines below.


   The interface responds to I/O instructions as follows:

   Output Data Word format (OTA and OTB, Control flip-flop set):

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | L | -   -   -   -   - | E | A |       output data byte        | unpacked data
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |    first output data byte     |    second output data byte    | packed data
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     L = Last byte out
     E = Assert EOI with byte
     A = Assert ATN with byte


   Output Control Word format (OTA and OTB, Control flip-flop clear):

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | S | I | R | V | L | F | E | A | D | N | P | O | K | L | T | C | configuration
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     S = Enable SRQ interrupt request
     I = Assert IFC or enable IFC interrupt request
     R = Assert REN
     V = Enable input-register-is-valid (loaded) interrupt request
     L = Enable last-byte-out interrupt request
     F = Enable line feed terminator
     E = Assert EOI
     A = Assert ATN
     D = Diagnostic loopback
     N = Assert NRFD
     P = Parallel poll enable
     O = Odd number of bytes
     K = Packed data transfer
     L = Listen
     T = Talk
     C = Controller in Charge


   Input Data Word format (LIA and LIB, Control flip-flop set):

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | -   -   -   -   -   - | E | A |        input data byte        | unpacked data
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |     first input data byte     |    second input data byte     | packed data
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     E = EOI was asserted with byte
     A = ATN was asserted with byte


   Input Status Word format (LIA and LIB, Control flip-flop clear):

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | S | I | R | V | L | B | E | A | F | - | - | O | K | L | T | C | status
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     S = SRQ is asserted
     I = IFC has been asserted
     R = REN is asserted
     V = Input register is valid (has received data)
     L = Last byte has been sent
     B = Last byte has been received
     E = EOI is asserted
     A = ATN is asserted
     F = IFC timer is still active
     O = Odd number of bytes
     K = System controller
     L = Listen
     T = Talk
     C = Controller in Charge


   Implementation notes:

    1. Card switches S1-S7 set the interface address and default to all switches
       closed, providing a card bus address of 0.  The address may be changed
       with the SET <dev> ADDRESS=n command.  Only addresses 0-7 are supported,
       and the address may duplicate a device bus address without conflict, as
       the address is used only during the interface diagnostic when devices are
       disconnected.

    2. Card switch S8 defaults to open, enabling the card to be the system
       controller.  This cannot be changed by the user.

    3. Card jumper W1 (DCPC pacing) is in position B, allowing SRQ to assert
       continuously for maximum performance.  This currently cannot be changed
       by the user.
*/



#include "hp2100_defs.h"
#include "hp2100_io.h"
#include "hp2100_di.h"



/* Program constants */

#define SW8_SYSCTL          1                   /* card is always the system controller (switch 8 open) */

#define IFC_TIMEOUT         uS (100)            /* IFC timer active for 100 microseconds */

#define CONTROLLER          30                  /* dummy bus address for the interface */


/* Character constants */

#define LF                  '\012'              /* line feed */


/* Interface Control Word.

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | S | I | R | V | L | F | E | A | D | N | P | O | K | L | T | C |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define CN_SRQ              0100000u            /* enable service request interrupt */
#define CN_IFC              0040000u            /* assert IFC or enable IFC interrupt */
#define CN_REN              0020000u            /* assert remote enable */
#define CN_IRL              0010000u            /* enable input-register-loaded interrupt */
#define CN_LBO              0004000u            /* enable last-byte-out interrupt */
#define CN_LF               0002000u            /* enable line feed terminator */
#define CN_EOI              0001000u            /* assert end or identify */
#define CN_ATN              0000400u            /* assert attention */
#define CN_DIAG             0000200u            /* diagnostic loopback */
#define CN_NRFD             0000100u            /* assert not ready for data */
#define CN_PPE              0000040u            /* parallel poll enable */
#define CN_ODD              0000020u            /* odd number of bytes */
#define CN_PACK             0000010u            /* packed data transfer */
#define CN_LSTN             0000004u            /* listen */
#define CN_TALK             0000002u            /* talk */
#define CN_CIC              0000001u            /* controller in charge */

#define CN_POLL             (CN_PPE | CN_CIC)   /* must be CIC in order to conduct poll */

static const BITSET_NAME control_names [] = {   /* interface control word names */
    "SRQ",                                      /*   bit 15 */
    "IFC",                                      /*   bit 14 */
    "REN",                                      /*   bit 13 */
    "IRL",                                      /*   bit 12 */
    "LBO",                                      /*   bit 11 */
    "LF",                                       /*   bit 10 */
    "EOI",                                      /*   bit  9 */
    "ATN",                                      /*   bit  8 */
    "Diagnostic",                               /*   bit  7 */
    "NRFD",                                     /*   bit  6 */
    "PPE",                                      /*   bit  5 */
    "Odd Count",                                /*   bit  4 */
    "Packed",                                   /*   bit  3 */
    "Listen",                                   /*   bit  2 */
    "Talk",                                     /*   bit  1 */
    "CIC"                                       /*   bit  0 */
    };

static const BITSET_FORMAT control_format =     /* names, offset, direction, alternates, bar */
    { FMT_INIT (control_names, 0, msb_first, no_alt, no_bar) };


/* Interface status word.

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | S | I | R | V | L | B | E | A | F | - | - | O | K | L | T | C |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define ST_SRQ              0100000u            /* SRQ bus state */
#define ST_IFCFF            0040000u            /* IFC flip-flop state */
#define ST_REN              0020000u            /* REN bus state */
#define ST_IRL              0010000u            /* input register loaded */
#define ST_LBO              0004000u            /* last byte out */
#define ST_LBI              0002000u            /* last byte in */
#define ST_EOI              0001000u            /* EOI bus state */
#define ST_ATN              0000400u            /* ATN bus state */
#define ST_IFC              0000200u            /* IFC asserted */
#define ST_ODD              0000020u            /* odd number of bytes */
#define ST_SYSCTL           0000010u            /* system controller */
#define ST_LSTN             0000004u            /* listener */
#define ST_TALK             0000002u            /* talker */
#define ST_CIC              0000001u            /* controller in charge */

static const BITSET_NAME status_names [] = {    /* interface status word names */
    "SRQ",                                      /*   bit 15 */
    "IFC Seen",                                 /*   bit 14 */
    "REN",                                      /*   bit 13 */
    "IRL",                                      /*   bit 12 */
    "LBO",                                      /*   bit 11 */
    "LBI",                                      /*   bit 10 */
    "EOI",                                      /*   bit  9 */
    "ATN",                                      /*   bit  8 */
    "IFC",                                      /*   bit  7 */
    NULL,                                       /*   bit  6 */
    NULL,                                       /*   bit  5 */
    "Odd Count",                                /*   bit  4 */
    "SYSCTL",                                   /*   bit  3 */
    "Listen",                                   /*   bit  2 */
    "Talk",                                     /*   bit  1 */
    "CIC"                                       /*   bit  0 */
    };

static const BITSET_FORMAT status_format =      /* names, offset, direction, alternates, bar */
    { FMT_INIT (status_names, 0, msb_first, no_alt, no_bar) };


/* Interface data words.

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | L | -   -   -   -   - | E | A |       output data byte        |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | -   -   -   -   -   - | E | A |        input data byte        |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |      first I/O data byte      |     second I/O data byte      |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define DATA_LBO            0100000u            /* last byte out */
#define DATA_EOI            0001000u            /* end or identify */
#define DATA_ATN            0000400u            /* attention */

static const BITSET_NAME data_names [] = {      /* data send and receive names */
    "Last Byte Out",                            /*   bit 15 */
    NULL,                                       /*   bit 14 */
    NULL,                                       /*   bit 13 */
    NULL,                                       /*   bit 12 */
    NULL,                                       /*   bit 11 */
    NULL,                                       /*   bit 10 */
    "EOI",                                      /*   bit  9 */
    "ATN"                                       /*   bit  8 */
    };

static const BITSET_FORMAT data_format =        /* names, offset, direction, alternates, bar */
    { FMT_INIT (data_names, 8, msb_first, no_alt, append_bar) };


/* First-in first-out memory.

   The interface card FIFO is 20 bits wide and 16 words deep.  It accepts 16
   bits of data plus four "tag" bits that associate actions with the data words.
   The tag bits are necessary to ensure that the actions are coordinated with
   the words when they emerge after falling through the FIFO.  The FIFO
   accommodates both single-byte and two-byte packed values.  It is implemented
   as a circular array of 32-bit unsigned integers organized as follows:

      19  18  17  16  15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | L | T | E | A |        upper data byte        |        lower data byte        |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     L = this word contains the last byte received
     T = this word ends the data transfer
     E = assert EOI (end or identify) with the byte
     A = assert ATN (attention) with the byte


   Implementation notes:

    1. EOI and ATN align with their counterparts in the HP-IB bus representation
       when right-shifted by eight bits.
*/

#define TAG_SHIFT           8                   /* left shift count to align bus ATN, EOI with FIFO tag */

#define TO_TAG(b)           ((b) << TAG_SHIFT)
#define TO_BUS(t)           ((t) >> TAG_SHIFT)

#define TAG_LBR             0002000000u         /* bit 19 = last byte received */
#define TAG_EDT             0001000000u         /* bit 18 = end of data transfer */
#define TAG_EOI             0000400000u         /* bit 17 = end or identify */
#define TAG_ATN             0000200000u         /* bit 16 = attention */

#define TAG_MASK            (TAG_LBR | TAG_EDT | TAG_EOI | TAG_ATN)

static const BITSET_NAME tag_names [] = {       /* Bus tag names */
    "ATN",                                      /*   bit 16 */
    "EOI",                                      /*   bit 17 */
    "EDT",                                      /*   bit 18 */
    "LBR"                                       /*   bit 19 */
    };

static const BITSET_FORMAT tag_format =         /* names, offset, direction, alternates, bar */
    { FMT_INIT (tag_names, 16, lsb_first, no_alt, no_bar) };

/* FIFO access modes */

typedef enum {
    bus_access,                                 /* per-byte access */
    cpu_access,                                 /* per-word access */
    diag_access                                 /* mixed access */
    } FIFO_ACCESS;

/* FIFO access tests */

#define FIFO_EMPTY          (di_card->fifo_count == 0)          /* FIFO empty test */
#define FIFO_FULL           (di_card->fifo_count == FIFO_SIZE)  /* FIFO full test */


/* Bus signals */

static const BITSET_NAME bus_names [] = {       /* Bus state names */
    "\1ATN\0   ",                               /*   bit  8 = attention */
    "\1EOI\0   ",                               /*   bit  9 = end or identify */
    NULL,                                       /*   bit 10 = not data accepted */
    NULL,                                       /*   bit 11 = data available */
    "\1NRFD\0    ",                             /*   bit 12 = not ready for data */
    NULL,                                       /*   bit 13 = remote enable */
    "\1IFC\0   ",                               /*   bit 14 = interface clear */
    NULL                                        /*   bit 15 = service request */
    };

static const BITSET_FORMAT bus_format =         /* names, offset, direction, alternates, bar */
    { FMT_INIT (bus_names, 8, lsb_first, has_alt, no_bar) };


/* Bus handlers table.

   The bus handlers table contains entries for the bus initializer, acceptor,
   and responder routines, as well as a pointer to the controlling device, for
   each protocol module.  RTE restricts each interface card to a single
   protocol, and the table entry corresponding to the interface card ID is used
   to call the protocol handler routines.  The associated device pointer is used
   for tracing as well as determining device flag states.

   A protocol's bus initializer is called during I/O initialization, which is
   invoked from the instruction execution prelude.  If the initializer returns
   FALSE, execution is inhibited.  If protocol initialization is not used, the
   field may be set to NULL.


   Implementation notes:

    1. The table nominally contains three entries, one for each protocol
       handler.  However, the Amigo tape (MA) entry is currently reserved for
       future use.
*/

extern DEVICE da_dev;
extern DEVICE dc_dev;
/* extern DEVICE ma_dev; */

typedef struct {                                /* bus protocol handler */
    DEVICE       *dptr;                         /*   a pointer to the controlling device */
    ACCEPTOR     *accept;                       /*   a pointer to the bus acceptor routine */
    RESPONDER    *respond;                      /*   a pointer to the bus responder routine */
    INITIALIZER  *initialize;                   /*   a pointer to the bus initializer routine */
    } BUS_HANDLER;

static BUS_HANDLER bus [Card_Count] = {         /* the bus handler table, indexed by CARD_ID */
/*    Device   Acceptor        Responder        Initializer        */
/*    -------  --------------  ---------------  ------------------ */
    { &da_dev, &da_bus_accept, &da_bus_respond, &da_bus_initialize },
    { &dc_dev, &dc_bus_accept, &dc_bus_respond, &dc_bus_initialize }
/*  { &ma_dev, &ma_bus_accept, &ma_bus_respond, &ma_bus_initialize } */
    };


/* Interface global state variables */

DI_STATE di [Card_Count];                       /* per-card state */


/* Interface local bus routines */

static t_bool   di_bus_accept  (CARD_ID card, BUS_CABLE signals);
static BUS_DATA di_bus_respond (CARD_ID card, BUS_CNTL  new_cntl);
static void     conduct_poll   (CARD_ID card, BUS_DATA  poll_response);

/* Interface local utility routines */

static void    master_reset (CARD_ID card);
static void    update_state (CARD_ID card);
static void    fifo_load    (CARD_ID card, FIFO_ACCESS access, HP_WORD data);
static HP_WORD fifo_unload  (CARD_ID card, FIFO_ACCESS access);


/* Interface global SCP data structures */

DEBTAB di_deb [] = {                            /* debugging trace list */
    { "CMD",   TRACE_CMD   },                   /*   trace interface or controller commands */
    { "INCO",  TRACE_INCO  },                   /*   trace interface or controller command initiations and completions */
    { "CSRW",  TRACE_CSRW  },                   /*   trace interface control, status, read, and write actions */
    { "STATE", TRACE_STATE },                   /*   trace state changes */
    { "SERV",  TRACE_SERV  },                   /*   trace unit service scheduling calls and entries */
    { "XFER",  TRACE_XFER  },                   /*   trace data transmissions and receptions */
    { "BUF",   TRACE_FIFO  },                   /*   trace FIFO buffer reads and writes */
    { "IOBUS", TRACE_IOBUS },                   /*   trace interface I/O bus signals and data words */
    { NULL,    0           }
    };



/* Interface global SCP support routines */


/* Disc interface.

   The disc interface is installed on the I/O bus and receives I/O commands from
   the CPU and DMA/DCPC channels.  In simulation, the asserted signals on the
   bus are represented as bits in the inbound signal set.  Each signal is
   processed sequentially in numerical order.

   The interface card has two input and two output registers.  The Input Data
   Register and Output Data Register are addressed when the Control flip-flop is
   set.  The Status Word Register and the Control Word Register are addressed
   when the Control flip-flop is clear.  The card has the usual Control, Flag
   Buffer, and Flag flip-flops.  A separate SRQ flip-flop is used to allow the
   full DCPC transfer rate.

   In hardware, the presence of the card FIFO, which is necessary to obtain full
   DCPC bandwidth, implies a delay between CPU actions, such as writing the last
   word of a data transfer, and device actions, such as accepting the last word
   of a disc write.  Four flip-flops are used to account for the FIFO delays:

     - EDT (End of Data Transfer)
     - LBO (Last Byte Out)
     - LBI (Last Byte In)
     - EOR (End of Record)

   The EDT signal at the interface indicates that the final data word of a
   transfer is being written to the FIFO.  The flip-flop is set by the EDT
   backplane signal when the last cycle of a DCPC transfer is executing, or
   during programmed output transfers when CLF does not accompany IOO in packed
   mode, or when bit 15 of the data word is set in unpacked mode.  It remains
   set until it is cleared by a master reset.  The output of the EDT flip-flop
   drives the EDT tag input of the FIFO.

   The LBO signal indicates that the final data byte of a transfer has been
   sourced to the HP-IB.  The flip-flop is set when the last byte of the entry
   tagged with EDT has been unloaded from the FIFO.  It is cleared by a master
   reset or when an entry not tagged with EDT is unloaded.  The output of the
   LBO flip-flop drives the LBO bit in the Status Word Register.

   The LBI signal indicates that the final byte of an input transfer has been
   accepted from the bus.  The flip-flop is set when a byte tagged with EOI is
   received and the EOI bit in the control register is set, or a line-feed byte
   is received and the LF bit in the control register is set.  It is cleared by
   a master reset or when neither of these conditions is true.  The input of the
   LBI flip-flop also drives the LBR (last byte received) tag input of the FIFO,
   and the output of the flip-flop drives the LBI bit in the Status Word
   Register.

   The EOR signal indicates that the final data word of a transfer is available
   in the Input Data Register.  The flip-flop is set when the last byte of the
   entry tagged with LBR has been unloaded from the FIFO and written to the IDR.
   It is cleared by a master reset or when an entry not tagged with LBR is
   unloaded and written to the IDR.  The output of the EOR flip-flop sets the
   Flag Buffer flip-flop when the IDR is unloaded.


   Implementation notes:

    1. In hardware, the Status Word Register consists of individual flip-flops
       and status signals that are enabled onto the I/O backplane.  In
       simulation, the individual status values are collected into a Status Word
       Register variable, and the Output Data Register does not exist (output
       data is written directly to the FIFO buffer).

    2. The DIAG, T, and L control bits enable a data loopback path on the card.
       An IOO issued to the card unloads a word from the FIFO and then loads the
       lower byte back into both bytes of the FIFO.  The data word output with
       the IOO instruction is not used.

       In hardware, IOO triggers the FIFO unload and reload; T and L are
       required only for the loopback path.  If L is not asserted, then the FIFO
       is loaded with 177777 due to the floating bus.  If L is asserted and T is
       not, then the FIFO is loaded with 000000 due to pullups on the DIO lines.
       In simulation, we look only for DIAG and assume that T/L are set
       properly, i.e., unloaded data is reloaded.

    3. In hardware, the HP-IB SRQ and NRFD lines are open-collector and may be
       driven simultaneously from several bus devices.  Simulating this fully
       would require keeping the state of the lines for each device and deriving
       the common bus signals from the logical OR of the state values.
       Fortunately, some simplifications are possible.

       The DI asserts SRQ only if control word bit 15 is 1 and bit 0 is 0.
       Other bit combinations deny SRQ; as neither the Amigo nor CS/80 protocols
       use SRQ and serial polls, there will be no other driver.

       In hardware, every listener drives NRFD, but in practice there is only
       one listener at a time.  When the card is the listener, it asserts NRFD
       if the FIFO becomes full.  In simulation, we assert NRFD on the bus if
       NRFD is set in the control register, or we are listening and the FIFO is
       full.  We deny NRFD if NRFD had been set in the control register but is
       no longer, or if we had been a listener but are no longer.  That is, we
       assume that if we have forced NRFD or set it as a listener, then no one
       else will be asserting NRFD, so it's safe for us to deny NRFD when the
       override is removed or we are no longer a listener.

       We also deny NRFD when a CRS is issued if NRFD had been explicitly
       requested or the card had been listening.  The rationale is the same:
       only a listener can assert NRFD, so if we were listening, it's safe to
       deny it, because only we could have set it.

    4. In hardware, the IRL, LBO, LBI, and IFC status bits are driven by
       corresponding flip-flops.  In simulation, the status bits themselves hold
       the equivalent states and are set and cleared as indicated.

    5. The card state must be updated during status read (IOI) processing
       because the 12992L boot loader ROM for the 7974A tape drive watches the
       IFC line to determine when IFC assertion ends.

    6. DCPC performance is optimized by recognizing that the normal cases (an
       input that empties the FIFO or an output that fills the FIFO) do not
       alter the card state, and so the usual update_state call may be omitted.

    7. For a Status Word Register read, the Talk and Listen bits from the
       control register are ORed, rather than set, into the status register, as
       these bits can also be set by receiving MTA or MLA from the bus.

    8. The gcc compiler (at least as of version 4.6.2) does not optimize
       repeated use of array-of-structures accesses.  Instead, it recalculates
       the index each time, even though the index is a constant within the
       function.  To avoid this performance penalty, we use a pointer to the
       selected DI_STATE structure.  Note that VC++ 2008 does perform this
       optimization.

    9. SRQ signal assertion via the SRQ flip-flop is qualified by the Control
       flip-flop, so that DCPC transfers occur only when the card is in data
       mode (Control is set) and the Input Data and Output Data Registers are
       addressed.  If the card is in configuration mode (Control is clear), the
       SRQ flip-flop output is blocked until data mode is reestablished.
*/

SIGNALS_VALUE di_interface (const DIB *dibptr, INBOUND_SET inbound_signals, HP_WORD inbound_value)
{
const CARD_ID card = (CARD_ID) (dibptr->card_index);    /* the ID number of the selected interface card */
DI_STATE * const di_card = &di [card];                  /* a constant pointer to the selected card state */

BUS_CNTL       assert, deny;                            /* the new bus control states */
INBOUND_SIGNAL signal;
INBOUND_SET    working_set = inbound_signals;
SIGNALS_VALUE  outbound    = { ioNONE, 0 };
t_bool         irq_enabled = FALSE;
t_bool         update_required = TRUE;                  /* TRUE if CLF must update the card state */

while (working_set) {                                   /* while signals remain */
    signal = IONEXTSIG (working_set);                   /*   isolate the next signal */

    switch (signal) {                                   /* dispatch the I/O signal */

        case ioCLF:                                     /* Clear Flag flip-flop */
            di_card->flag_buffer = CLEAR;               /* clear the flag buffer */
            di_card->flag        = CLEAR;               /*   and flag flip-flops */

            if (update_required)                        /* if the card state has changed */
                update_state (card);                    /*   then update the state */
            break;


        case ioSTF:                                     /* Set Flag flip-flop */
            di_card->flag_buffer = SET;                 /* set the flag buffer flip-flop */
            break;


        case ioENF:                                     /* Enable Flag */
            if (di_card->flag_buffer == SET)            /* if the flag buffer flip-flop is set */
                di_card->flag = SET;                    /*   then set the flag flip-flop */
            break;


        case ioSFC:                                     /* Skip if Flag is Clear */
            if (di_card->flag == CLEAR)                 /* if the flag flip-flop is clear */
                outbound.signals |= ioSKF;              /*   then assert the Skip on Flag signal */
            break;


        case ioSFS:                                     /* Skip if Flag is Set */
            if (di_card->flag == SET)                   /* if the flag flip-flop is set */
                outbound.signals |= ioSKF;              /*   then assert the Skip on Flag signal */
            break;


        case ioIOI:                                             /* I/O data input */
            if (di_card->control == SET) {                      /* if the card is in data mode */
                outbound.value = di_card->input_data_register;  /*   then read the input data register */
                di_card->status_register &= ~ST_IRL;            /*     and clear the input register loaded status */

                if (FIFO_EMPTY && di_card->eor == CLEAR) {  /* if the is FIFO empty and end-of-record is not seen */
                    di_card->srq = CLEAR;                   /*   then clear SRQ */
                    update_required = FALSE;                /*     and the card state does not change */
                    }

                if (di_card->cntl_register & CN_PACK)       /* if this is a packed transfer */
                    tpprintf (bus [card].dptr, TRACE_CSRW, "Data %06o received from bus\n",
                              outbound.value);
                else
                    tpprintf (bus [card].dptr, TRACE_CSRW, "Data %s%03o received from bus\n",
                              fmt_bitset (outbound.value, data_format),
                              LOWER_BYTE (outbound.value));
                }

            else {                                              /* otherwise the card is in status mode */
                di_card->status_register &=                     /*   so clear the values to be computed */
                  ST_IRL | ST_LBO | ST_LBI | ST_IFC | ST_IFCFF; /*     preserving those set elsewhere */

                di_card->status_register |=                 /* merge T/L/C status from the control register */
                  di_card->cntl_register & (CN_CIC | CN_TALK | CN_LSTN);

                if (SW8_SYSCTL)                             /* if SW8 is set, */
                    di_card->status_register |= ST_SYSCTL;  /*   then the card is the system controller */

                if (di_card->ibp == lower)                  /* if lower byte input is next */
                    di_card->status_register |= ST_ODD;     /*   then the last transfer was odd */

                di_card->status_register |=                 /* merge the bus status bits */
                  di_card->bus_cntl & (Bus_SRQ | Bus_IFC    /*   from the corresponding bus control lines */
                                     | Bus_REN | Bus_EOI | Bus_ATN);

                outbound.value = di_card->status_register;  /* return the status word */

                tpprintf (bus [card].dptr, TRACE_CSRW, "Status is %s\n",
                          fmt_bitset (outbound.value, status_format));
                }

            if (update_required                         /* if an update is required */
              && not (inbound_signals & ioCLF))         /*   but CLF is not present */
                update_state (card);                    /*     then update the state, else ioCLF will update it */
            break;


        case ioIOO:                                     /* I/O data output */
            if (di_card->control == SET) {              /* if the card is in data mode */
                if (di_card->cntl_register & CN_PACK)   /*   then if this is a packed transfer */
                    tpprintf (bus [card].dptr, TRACE_CSRW, "Data %06o sent to bus\n",
                              inbound_value);
                else
                    tpprintf (bus [card].dptr, TRACE_CSRW, "Data %s%03o sent to bus\n",
                              fmt_bitset (inbound_value, data_format),
                              LOWER_BYTE (inbound_value));

                if (inbound_signals & ioEDT)            /* if a DCPC transfer has ended */
                    di_card->edt = SET;                 /*   then set the EDT flip-flop */

                else if (di_card->cntl_register & CN_PACK) {    /* otherwise if this is a packed transfer */
                    if (not (inbound_signals & ioCLF))          /*   and CLF is not given */
                        di_card->edt = SET;                     /*     then set the EDT flip-flop */
                    }

                else                                    /* otherwise it's an unpacked transfer */
                    if (inbound_value & DATA_LBO)       /*   so if this is the last byte out */
                        di_card->edt = SET;             /*     then set the EDT flip-flop */

                if (di_card->cntl_register & CN_DIAG) {                 /* if diagnostic mode is active */
                    inbound_value = fifo_unload (card, diag_access);    /*   then unload data from the FIFO */
                    fifo_load (card, diag_access, inbound_value);       /*     and load it back in */
                    }

                else {                                              /* otherwise the bus cable is connected */
                    fifo_load (card, cpu_access, inbound_value);    /*   so load the data word into the FIFO */

                    if (FIFO_FULL && (di_card->bus_cntl & Bus_NRFD)) {  /* if the FIFO is full and NRFD is asserted */
                        di_card->srq = CLEAR;                           /*   then clear SRQ */
                        update_required = FALSE;                        /*     and the card state does not change */
                        }
                    }
                }

            else {                                      /* otherwise the card is in control mode */
                tpprintf (bus [card].dptr, TRACE_CSRW, "Control is %s\n",
                          fmt_bitset (inbound_value, control_format));

                assert = BUS_NONE;                      /* initialize bus control assertions */
                deny = BUS_NONE;                        /*   and denials */

                if (not (inbound_value & CN_PACK))          /* if unpacked mode is selected */
                    di_card->ibp = di_card->obp = lower;    /*   then restrict access to the lower byte */

                if (inbound_value & CN_TALK) {                  /* if the ATN and EOI outputs are enabled */
                    if ((inbound_value & CN_POLL) == CN_POLL)   /*   then if parallel poll is enabled */
                        assert = BUS_PPOLL;                     /*     then request a parallel poll */

                    else if ((di_card->cntl_register & CN_POLL) == CN_POLL) /* otherwise if a poll had been enabled */
                        deny = BUS_PPOLL;                                   /*   then end the parallel poll */

                    else if ((inbound_value & (CN_PACK | CN_CIC | CN_ATN))  /* otherwise if in packed mode */
                      == (CN_PACK | CN_CIC | CN_ATN))                       /*   and CIC and ATN is requested */
                        assert = Bus_ATN;                                   /*     then couple ATN to the bus */

                    else                                /* otherwise */
                        deny = Bus_ATN;                 /*   ATN is not driven */
                    }

                else                                    /* otherwise the card is not talking */
                    deny = Bus_ATN | Bus_EOI;           /*   so ATN and EOI are disabled */

                if (inbound_value & CN_NRFD)            /* if the card is not ready */
                    assert |= Bus_NRFD;                 /*   then assert NRFD on the bus */

                else if (di_card->cntl_register & CN_NRFD)  /* otherwise if NRFD was set but is not now */
                    deny |= Bus_NRFD;                       /*   then deny NRFD on the bus */

                if (FIFO_FULL)                          /* if the FIFO is full */
                    if (inbound_value & CN_LSTN)        /*   then if the card is listening */
                        assert |= Bus_NRFD;             /*     then assert NRFD to suspend data transfer */

                    else if (di_card->cntl_register & CN_LSTN)  /* otherwise if the card was listening but is not now */
                        deny |= Bus_NRFD;                       /*   then deny NRFD on the bus */

                if (SW8_SYSCTL) {                       /* if the card is the system controller */
                    if (inbound_value & CN_REN)         /*   then the REN */
                        assert |= Bus_REN;              /*     control output */
                    else                                /*       is coupled */
                        deny |= Bus_REN;                /*         to the bus */

                    if (inbound_value & CN_IFC) {       /* if IFC is set */
                        assert |= Bus_IFC;              /*   then assert IFC on the bus */

                        di_card->status_register =      /* clear listen and talk status */
                          di_card->status_register      /*   and set */
                          & ~(ST_LSTN | ST_TALK)        /*     the IFC signal and */
                          | ST_IFC | ST_IFCFF;          /*       IFC flip-flop status */

                        di_card->ifc_timer =            /* start the IFC timer by calculating */
                          sim_gtime () + IFC_TIMEOUT;   /*   the IFC stop time (now + 100 microseconds) */
                        }
                    }

                if ((inbound_value & (CN_SRQ | CN_CIC)) == CN_SRQ)  /* if SRQ but we are not the CIC */
                    assert |= Bus_SRQ;                              /*     then assert SRQ on the bus */
                else                                                /*   else */
                    deny |= Bus_SRQ;                                /*     deny SRQ on the bus */

                di_card->cntl_register = inbound_value;             /* save the control word */
                di_bus_control (card, CONTROLLER, assert, deny);    /*   and update the bus control state */
                }

            if (update_required                         /* if an update is required */
              && not (inbound_signals & ioCLF))         /*   but CLF is not present */
                update_state (card);                    /*     then update the state, else ioCLF will update it */
            break;


        case ioPOPIO:                                   /* Power-On Preset to I/O */
            di_card->flag_buffer = SET;                 /* set the flag buffer flip-flop */
            break;


        case ioCRS:                                     /* Control Reset */
            tpprintf (bus [card].dptr, TRACE_CSRW, "Control reset\n");

            di_card->status_register &= ~(ST_LSTN | ST_TALK);   /* unlisten and untalk the card */

            deny = Bus_SRQ | Bus_REN | Bus_ATN | Bus_EOI;       /* clear lines driven by the control register */

            if (di_card->cntl_register & (CN_NRFD | CN_LSTN))   /* if asserting NRFD or listening */
                deny |= Bus_NRFD;                               /*   then deny because we're clearing */

            di_card->cntl_register = 0;                 /* clear the control word register */
            di_card->control = CLEAR;                   /*   and the control flip-flop */
            di_card->srq = CLEAR;                       /*     and the SRQ flip-flop */

            master_reset (card);                        /* perform a master reset */

            di_bus_control (card, CONTROLLER, BUS_NONE, deny);  /* update the bus control state */
            update_state (card);                                /*   and the card state */
            break;


        case ioCLC:                                     /* Clear Control flip-flop */
            di_card->control = CLEAR;                   /* clear the control flip-flop */

            if (inbound_signals & ioCLF)                /* if ioCLF is given, */
                master_reset (card);                    /*   then do a master reset */
            break;                                      /*     (ioCLF will call update_state for us) */


        case ioSTC:                                     /* Set Control flip-flop */
            di_card->control = SET;                     /* set the control flip-flop */

            if (not (inbound_signals & ioCLF))          /* if CLF is not present, */
                update_state (card);                    /*   then update the state, else ioCLF will update it */
            break;


        case ioEDT:                                     /* End Data Transfer */
            tpprintf (bus [card].dptr, TRACE_CSRW, "DCPC transfer ended\n");
            break;


        case ioSIR:                                     /* Set Interrupt Request */
            if (di_card->control & di_card->flag)       /* if the control and flag flip-flops are set */
                outbound.signals |= cnVALID;            /*   then deny PRL */
            else                                        /* otherwise */
                outbound.signals |= cnPRL | cnVALID;    /*   conditionally assert PRL */

            if (di_card->control & di_card->flag        /* if the control and flag */
              & di_card->flag_buffer)                   /*   and flag buffer flip-flops are set */
                outbound.signals |= cnIRQ;              /*     then conditionally assert IRQ */

            if (di_card->control & di_card->srq)        /* if the control and srq flip-flops are set */
                outbound.signals |= ioSRQ;              /*   then assert SRQ */
            break;


        case ioIAK:                                     /* Interrupt Acknowledge */
            di_card->flag_buffer = CLEAR;               /* clear the flag buffer flip-flop */
            break;


        case ioIEN:                                     /* Interrupt Enable */
            irq_enabled = TRUE;                         /* permit IRQ to be asserted */
            break;


        case ioPRH:                                         /* Priority High */
            if (irq_enabled && outbound.signals & cnIRQ)    /* if IRQ is enabled and conditionally asserted */
                outbound.signals |= ioIRQ | ioFLG;          /*   then assert IRQ and FLG */

            if (not irq_enabled || outbound.signals & cnPRL)    /* if IRQ is disabled or PRL is conditionally asserted */
                outbound.signals |= ioPRL;                      /*   then assert it unconditionally */
            break;


        case ioPON:                                     /* not used by this interface */
            break;
        }

    IOCLEARSIG (working_set, signal);                   /* remove the current signal from the set */
    }                                                   /*   and continue until all signals are processed */

return outbound;                                        /* return the outbound signals and value */
}


/* Disc interface initializer.

   This routine is called from the instruction execution prelude to set up the
   internal state of the disc interface and its connected device controller
   prior to instruction execution.

   If the card is in DIAGNOSTIC mode, the routine searches for the other disc
   interface card in the same mode and, if found, connects the two cards with an
   HP-IB cable.  This permits the card-to-card test to be run.

   The routine also calls the card-specific initializer, if one is defined.  The
   initializer typically checks for conflicts among the devices on the bus.
*/

t_bool di_initializer (DEVICE *dptr, DIB *dibptr)
{
const CARD_ID card = (CARD_ID) (dibptr->card_index);    /* the ID number of the selected interface card */
CARD_ID       id;

di [card].other = Last_Card + 1;                        /* assume no cross-connected cards */

if (dptr->flags & DEV_DIAG)                             /* if the card is in diagnostic mode */
    for (id = First_Card; id <= Last_Card; id++)        /*   then look through the list of cards */
        if (id != card                                  /*     for another card */
          && bus [id].dptr != NULL                      /*       that is configured */
          && bus [id].dptr->flags & DEV_DIAG)           /*         for diagnostic mode */
            di [card].other = id;                       /*           and save its card ID */

if (bus [card].initialize != NULL)                      /* if a bus initializer is defined */
    return bus [card].initialize (card);                /*   then call it and return its status */
else                                                    /* otherwise */
    return TRUE;                                        /*   initialization succeeded */
}



/* Device reset routine.

   This routine is called by the specific protocol handlers for a RESET or RESET
   DA (DC, etc.) command.  It is the simulation equivalent of applying or
   cycling power on the set of bus-connected devices.

   As this is the interface reset routine, no action is taken on a normal reset.
   However, for a power-on reset, pointers to the FIFO and TAG simulation
   registers are saved to allow access to the "qptr" fields during FIFO loading
   and unloading.  This enables SCP to view the FIFO and TAG registers as
   circular queues, so that the bottom word of the FIFO is always displayed as
   FIFO[0], regardless of where it is in the actual array.
*/

t_stat di_reset (DEVICE *dptr)
{
DIB const *dibptr = (DIB *) dptr->ctxt;                 /* get the DIB pointer */
const CARD_ID card = (CARD_ID) (dibptr->card_index);    /*   and the card number */

if (sim_switches & SWMASK ('P')) {                      /* if this is a power-on reset */
    di [card].fifo_reg = find_reg ("FIFO", NULL, dptr); /*   then find the FIFO register entry */
    di [card].tag_reg  = find_reg ("FTAG", NULL, dptr); /*     and the the TAG register entry */

    if (di [card].fifo_reg == NULL                      /* if either register is not there */
      || di [card].tag_reg == NULL)                     /*   then report a programming error */
        return SCPE_IERR;                               /*     to the console */

    else                                                /* otherwise */
        di [card].fifo_reg->qptr =                      /*   reset the FIFO */
          di [card].tag_reg->qptr = 0;                  /*     and TAG bottom indices */

    di [card].fifo_count = 0;                           /* clear the FIFO */
    di [card].status_register = 0;                      /*   and the status word */
    di [card].bus_cntl = 0;                             /* deny the HP-IB control lines */
    di [card].ifc_timer = 0.0;                          /*   and clear the IFC timer */
    }

return SCPE_OK;
}


/* Set the interface bus address.

   This validation routine is entered with the "cptr" parameter pointing to the
   new address value and the "desc" parameter pointing to the DEVICE structure.
   The "uptr" and "value" parameters are not used.

   Bus addresses range from 0-7 and establish the card address on the bus for
   the diagnostic.  The bus address is not used when the card is connected to
   the bus devices, which have their own addresses.
*/

t_stat di_set_address (UNIT *uptr, int32 value, char *cptr, void *desc)
{
DEVICE *dptr = (DEVICE *) desc;                 /* a pointer to the DEVICE structure */
t_stat status;
uint32 new_address;

if (cptr == NULL)                                           /* if the address is not given */
    return SCPE_ARG;                                        /*   then report a missing argument */
else                                                        /* otherwise */
    new_address = (uint32) get_uint (cptr, 10, 7, &status); /*   parse the address value */

if (status == SCPE_OK)                                  /* if the parse succeeded  */
    dptr->flags = dptr->flags & ~DEV_BUSADR             /*   then store the new address in the device flags */
                    | TO_DEV_ADDR (new_address);

return status;                                          /* return the result of the parse */
}


/* Set the bus cable connection.

   This validation routine is called with the "value" parameter indicating
   whether the routine is connecting the bus cable to devices for normal use (0)
   or to another card for diagnostics (1).  The "desc" parameter points to the
   DEVICE structure, and the other parameters are not used.

   In normal use, the various tape and disc devices are connected together and
   to the disc interface card by HP-IB cables.  For the diagnostic, two disc
   interface cards are connected by a single cable.
*/

t_stat di_set_cable (UNIT *uptr, int32 value, char *cptr, void *desc)
{
DEVICE *dptr = (DEVICE *) desc;                 /* a pointer to the DEVICE structure */

if (value)                                              /* if the diagnostic cable is selected */
    dptr->flags |= DEV_DIAG;                            /*   then set the diagnostic flag */
else                                                    /* otherwise the peripheral cable is selected */
    dptr->flags &= ~DEV_DIAG;                           /*   so clear the diagnostic flag */

return SCPE_OK;                                         /* the routine always succeeds */
}


/* Show the interface bus address.

   This display routine is called to show the bus address for the interface
   card.  The output stream is passed in the "st" parameter, the "desc"
   parameter points to the DEVICE structure, and the other parameters are
   ignored.  The address is obtained from its location within the device flags.
*/

t_stat di_show_address (FILE *st, UNIT *uptr, int32 value, void *desc)
{
const DEVICE *dptr = (const DEVICE *) desc;     /* a constant pointer to the DEVICE structure */

fprintf (st, "address=%d", DEV_ADDR (dptr->flags));     /* print the address of the interface card */

return SCPE_OK;
}


/* Show the bus cable connection.

   This display routine is called to show the interface connection.  The output
   stream is passed in the "st" parameter, the "desc" parameter points to the
   DEVICE structure, and the other parameters are ignored.  The connection is
   obtained from its location within the device flags.
*/

t_stat di_show_cable (FILE *st, UNIT *uptr, int32 value, void *desc)
{
const DEVICE *dptr = (const DEVICE *) desc;     /* a constant pointer to the DEVICE structure */

if (dptr->flags & DEV_DIAG)                             /* if diagnostic mode is enabled */
    fputs ("diagnostic cable", st);                     /*   then two interfaces are connected together */
else                                                    /* otherwise */
    fputs ("HP-IB cable", st);                          /*   the interface is connected to the bus devices */

return SCPE_OK;
}



/* Disc interface global bus routines.

   In hardware, the HP-IB bus consists of eight control lines and eight data
   lines.  Signals are asserted on the control lines to establish communication
   between a source and one or more acceptors.  For commands, the source is
   always the controller (the 12821A card), and the acceptors are all of the
   connected devices.  For data, the source is the current talker, and the
   acceptors are one or more current listeners.  A three-wire interlocking
   handshake enables communication at the rate of the slowest of the multiple
   acceptors.  The controller conducts a parallel poll by asserting ATN and EOI
   together.  Devices whose parallel poll responses are enabled each assert one
   of the data lines to indicate that service is required.

   Functional simulation is partitioned into this module, which provides the
   12821A interface card, and the protocol modules, which provide the connected
   Amigo disc, CS/80 disc and tape, and Amigo tape devices.  Each protocol
   module connects to this module via a simulation of the HP-IB, which is
   modeled as the BUS_CABLE.  The cable consists of eight bits representing the
   control lines and 8 bits representing the data lines.

   Not all of the HP-IB control lines are simulated.  The DAV and NDAC handshake
   lines are never asserted.  Instead, they are simulated by a bus source
   function that implements the Source Handshake (SH) calling one or more bus
   acceptor functions that implement the Acceptor Handshake (AH).  SRQ and REN
   are asserted as directed by the system controller but are not otherwise used
   (no HP disc or tape devices assert SRQ or respond to REN).  IFC, ATN, EOI,
   and NRFD are asserted and tested by the interface and protocol modules.  In
   particular, asserting NRFD will hold off a pending data transmission until
   the signal is denied.

   The functions that simulate the bus interconnection between the CPU interface
   (this module) and the connected devices (the "da", "dc", "di", etc. modules)
   are:

                          Calling       Protocol
     Interface Function  Direction  Handler Function
     ------------------  ---------  ----------------
     di_bus_source          -->     **_bus_accept

     di_bus_control         -->     **_bus_respond

     di_bus_ready           <--     (various)


   The functions perform the following actions:

     t_bool di_bus_source
     --------------------

       Send the set of control signals and a data byte to the bus and perform an
       implied handshake.  Corresponds in hardware to driving the bus lines and
       executing the IEEE-488 SH (Source Handshake) interface function.  Returns
       TRUE if some device on the bus accepted the information (i.e., there were
       one or more listeners) and FALSE if not.  Called by the controller to
       command devices, and called by the current talker to send data to the
       current listener(s).

     t_bool **_bus_accept
     --------------------

       Receive the set of control signals and a data byte from the bus and
       complete the implied handshake.  Corresponds in hardware to receiving the
       bus lines and performing the IEEE-488 AH (Acceptor Handshake) interface
       function.  Returns TRUE if some device accepted the data and FALSE if
       not.  Called by di_bus_source to handshake between the controller and
       devices and between the talker and the listeners.  If ATN is asserted on
       the bus, the data byte is a command; otherwise, it is data.  If EOI is
       asserted for a data byte, it is the last byte of a transmission.

     void di_bus_control
     -------------------

       Change the control signals on the bus without performing an implied
       handshake.  Corresponds in hardware to performing the IEEE-488 SR
       (Service Request), RL (Remote Local), PP (Parallel Poll), and C
       (Controller) interface functions.  All connected devices on the bus are
       notified of the changes.  It is not necessary to call di_bus_control for
       changes to ATN or EOI that accompany a command or data byte; these
       signals are passed to a di_bus_source call.

     BUS_DATA **_bus_respond
     -----------------------

       Respond to control signal changes on the bus.  Corresponds in hardware to
       device state changes occurring by signal assertion or denial.  Called by
       di_bus_control to inform each connected device of a change in control
       state.  Returns the set of parallel poll responses for the connected
       devices (only relevant when ATN and EOI are asserted simultaneously to
       conduct a poll).

     t_bool di_bus_ready
     -------------------

       Check to see if the bus is ready for a data byte.  Returns TRUE if the
       ATN (Attention) and NRFD (Not Ready For Data) signals are both denied to
       indicate that the talker is ready to source data and the listener is
       ready to accept data, and FALSE otherwise.


   Implementation notes:

    1. If multiple protocols were supported on the same bus, these routines
       would need to call each of the protocol handlers for a given operation.
       However, as RTE permits only one protocol per bus, we can avoid that and
       call only the handler dedicated to the specific interface card.
*/


/* Source a byte to the bus.

   This routine is called to place a data byte on the bus and assert the
   specified set of control signals.  If the card is in diagnostic mode, which
   simulates two cards connected by an HP-IB cable, then the byte is sent to
   another card in the I/O card cage that is also in diagnostic mode and enabled
   to receive.  If the card is not in diagnostic mode, then the byte is sent to
   the bus acceptor of the protocol handler for the specified interface card.
   The return value indicates whether or not there were any acceptors on the
   bus.


   Implementation notes:

    1. If the responses from a previously conducted parallel poll were not
       cleared from the FIFO before enabling the card to transmit, the card will
       appear to conduct a new parallel poll because the FIFO tags cause ATN and
       EOI to be asserted.  This "fake" parallel poll is ignored (a real
       parallel poll does not source data onto the bus).
*/

t_bool di_bus_source (CARD_ID card, BUS_CABLE signals)
{
t_bool accepted = FALSE;

di [card].bus_cntl |= CNTL (signals);                   /* assert the specified bus signals */

signals |= di [card].bus_cntl;                          /* merge the new control signals with the data byte */

if (bus [card].dptr->flags & DEV_DIAG)                  /* if the card is in diagnostic mode */
    if (di [card].other <= Last_Card)                   /*   then if there is another card on the cable */
        accepted = di_bus_accept (di [card].other,      /*     then ask that card */
                                  signals);             /*       to accept the byte */
    else                                                /*   otherwise there is nobody */
        { ; }                                           /*     on the other end of the cable */

else if ((signals & BUS_PPOLL) != BUS_PPOLL) {          /* otherwise if this is not a "fake" poll */
    accepted = di_bus_accept (card, signals);           /*   then call the acceptor for the card */

    accepted |= bus [card].accept (card, signals);      /* call the acceptor for the bus devices */
    }

if (not accepted)
    tpprintf_bus (bus [card].dptr, signals, "No bus acceptors\n");

return accepted;                                        /* return the acceptance state */
}


/* Assert or deny control on the bus.

   This routine is called by the controller to assert or deny the HP-IB control
   lines on the bus connected to the specified card.  It is also called by the
   talker and listeners to assert or deny NRFD as needed to pace the data
   transmission.  Separate sets of signals to assert and deny are provided.

   If the bus state after modification did not change, and a parallel poll is
   not in progress, then the routine returns with no further action.  Otherwise,
   if the card is in diagnostic mode, then notification of the bus change is
   sent to another card in the I/O card cage that is also in diagnostic mode.

   If the card is not in diagnostic mode, then the set of control lines that
   are changing is checked to determine whether notification is necessary.  If
   not, then the change is not broadcast to improve performance.  However, if
   notification is required, then the bus responder of the protocol handler for
   the specified interface card is informed of the change.


   Implementation notes:

    1. If a signal is asserted and denied in the same call, the assertion takes
       precedence.

    2. Asserting ATN and EOI will conduct a parallel poll.  The bus responder
       returns the set of poll responses for the devices it controls.

    3. Of the sixteen potential control line state changes, only IFC assertion
       and ATN and NRFD denial must be broadcast to devices.  Asserting IFC
       unaddresses all devices, and denying ATN or NRFD allows a waiting talker
       to source a data byte to the bus.  Devices do not act upon the remaining
       thirteen state changes, and a considerable performance improvement is
       obtained by omitting the notification calls.

    4. All control line state notifications are sent in diagnostic mode, as the
       responses of the other card are specifically tested by the diagnostic.

    5. An active parallel poll must be broadcast to devices, regardless of
       whether it represents a change in the control line states.  This is
       because connected devices might change their poll response during the
       time a poll is active.  In particular, RTE-6/VM driver DVM33 conducts
       back-to-back parallel polls and expects that data will be loaded into the
       FIFO and trigger an Input Register Loaded interrupt.  This does not need
       to be done in diagnostic mode, as poll response changes can be initiated
       only by changing the card's control register.
*/

void di_bus_control (CARD_ID card, uint32 address, BUS_CNTL assert, BUS_CNTL deny)
{
t_bool   parallel_poll;
BUS_CNTL new_state, new_assertions, new_denials;
BUS_DATA response = 0;

if (address != CONTROLLER) {                            /* if the caller is not the controller */
    assert &= ~BUS_CNTLR;                               /*   then remove the signals */
    deny   &= ~BUS_CNTLR;                               /*     that devices are not allowed to change */
    }

new_state = di [card].bus_cntl & ~deny | assert;        /* set up the new control state */

parallel_poll = ((new_state & BUS_PPOLL) == BUS_PPOLL); /* TRUE if a parallel poll is in progress */

if (new_state != di [card].bus_cntl) {                  /* if the control state changed */
    new_assertions = ~di [card].bus_cntl & assert;      /*   then get the changing assertions */
    new_denials    =  di [card].bus_cntl & deny;        /*     and the changing denials */

    di [card].bus_cntl = new_state;                     /* establish the new control state */

    tpprintf_bus (bus [card].dptr, new_state, "Signal change (%s %u)\n",
                  (address == CONTROLLER ? "card" : "device"),
                  (address == CONTROLLER ? card : address));

    if (bus [card].dptr->flags & DEV_DIAG)              /* if the card is in diagnostic mode */
        if (di [card].other <= Last_Card)               /*   then if there is another card on the cable */
            response = di_bus_respond (di [card].other, /*     then notify that card */
                                       new_state);      /*       of the new control state */
        else                                            /*   otherwise there is nobody */
            { ; }                                       /*     on the other end of the cable */

    else {                                              /* otherwise the card connected to the bus */
        update_state (card);                            /*   so update it for the new control state */

        if (parallel_poll                                       /* if a parallel poll is active */
          || new_assertions & (Bus_IFC)                         /*   or there are changed signals */
          || new_denials & (Bus_ATN | Bus_NRFD))                /*     in the set that must be broadcast */
            response = bus [card].respond (card, new_state);    /*       then call the bus device responder */
        }
    }

else if (parallel_poll)                                 /* otherwise if a parallel poll is active */
    response = bus [card].respond (card, new_state);    /*   then call the bus device responder */

if (parallel_poll)                                      /* if a poll is active */
    conduct_poll (card, response);                      /*   then conduct it */

return;
}


/* Test if the bus is ready for data.

   This routine checks to see the bus is ready for a data byte.  Returns TRUE if
   the ATN (Attention) and NRFD (Not Ready For Data) signals are both denied,
   and FALSE otherwise.

   When a device is addressed to talk, it cannot send the first data byte until
   the ATN that accompanies the Talk command is denied by the controller.  In
   hardware, this corresponds to the transition from TADS (talker addressed
   state) to TACS (talker active state).  To then source a byte to the bus, NRFD
   must be denied, corresponding to the transition from SDYS (source delay
   state) to STRS (source transfer state).  A listener may hold off the talker
   by asserting NRFD until it is ready to receive.

   A device that wants to talk can call this routine to determine if the bus
   state permits it to source a byte.  If it returns FALSE, the device must wait
   until it receives notification of the bus state change via its bus responder
   function.
*/

t_bool di_bus_ready (CARD_ID card)
{
return not (di [card].bus_cntl & (Bus_ATN | Bus_NRFD)); /* return TRUE if a byte may be sourced */
}


/* Format a bus data transfer.

   This routine is called to write a line to the trace log for a bus transfer.
   The format employed is patterned after a logic analyzer, e.g.:

     DIO | ATN | EOI | NRFD | IFC | HP-IB Mnemonic
     94H | ATN |     |      |     | Universal clear
     80H | ATN | EOI |      |     | Parallel poll
     20H |     |     |      |     | Data (device 0 accepts)

   ...showing the data and signal state and a translation of the HP-IB command.

   On entry, "dptr" points to the device issuing the trace, "signals" is the set
   of bus signals and data present on the HP-IB cable, and the remaining
   parameters consist of the format string that provides the HP-IB Mnemonic part
   of the line, and any required parameters to satisfy the format string.  The
   trace is written for the TRACE_XFER category.


   Implementation notes:

    1. This routine could be expressed as a macro:

         #define di_bus_trace(dptr, signals, format, ...) \
             tpprintf (dptr, TRACE_XFER, "%02XH | %s | " format, \
                       DATA (signals), fmt_bitset (signals, bus_format), __VA_ARGS__);

       However, this would permit calls of the form:

         di_bus_trace (dptr, signals, "format %u", 0);

       ...but not of the form:

         di_bus_trace (dptr, signals, "format");

       The latter produces "error: expected expression before ')' token" in the
       macro expansion.  So we must perform the format concatenation in a
       routine.

    2. Ideally, "print_header" would be reset each time a new debug log file is
       defined.  But this would require hooking the SET CONSOLE command, just to
       ensure that a header is printed at the start of each file.  An
       alternative is to print the header each time execution starts, but that
       prints multiple headers if tracing is enabled while REPLYs are being
       used.  The implementation chosen prints the header once per simulation
       run.
*/

#define FORMAT_SIZE         1024                /* sufficiently large to accommodate all format strings */

void di_bus_trace (DEVICE *dptr, BUS_CABLE signals, ...)
{
const  DIB     *dibptr = (DIB *) dptr->ctxt;
const  CARD_ID card = (CARD_ID) (dibptr->card_index);
const  char    format [] = "%02XH | %s | ";
static t_bool  header_printed [Card_Count] = { FALSE };
va_list        argptr;
int            index;
char           *fptr, buffer [FORMAT_SIZE];

if (not header_printed [card]) {                        /* if the header for this card has not been printed */
    hp_trace (dptr, TRACE_XFER, "DIO | ATN | EOI | NRFD | IFC | HP-IB Mnemonic\n");
    header_printed [card] = TRUE;                       /* mark it as already printed */
    }

index = sprintf (buffer, format, DATA (signals),        /* format the DIO value and bus signals */
                 fmt_bitset (signals, bus_format));     /*   into a string buffer */

va_start (argptr, signals);                             /* set up the argument list */

fptr = va_arg (argptr, char *);                         /* get the HP-IB Mnemonic format string parameter */

vsprintf (buffer + index, fptr, argptr);                /* format and append it to the buffer */

va_end (argptr);                                        /* clean up the argument list */

hp_trace (dptr, TRACE_XFER, buffer);                    /* call the trace routine to write the line */

return;
}



/* Interface local bus routines */


/* Conduct a parallel poll on the bus.

   A controller asserting ATN and EOI simultaneously on the bus is conducting a
   parallel poll.  In hardware, each device whose poll response is enabled
   asserts the data line corresponding to its bus address.  The controller
   terminates the poll by denying ATN and EOI.

   Setting the CIC (controller in charge) and PPE (parallel poll enable) bits in
   the Control Word Register direct the disc interface to conduct a poll.
   Setting PPE without CIC enables the poll response for the interface.

   In the diagnostic mode, one card is set to conduct the poll, and the other is
   set to respond to it.  In the normal mode, connected devices have set or
   cleared their respective poll responses before this routine is called.


   Implementation notes:

    1. The card hardware fills the upper and lower bytes of the FIFO with the
       response byte.  In simulation, we use the diag_access mode to do the same
       thing (diagnostic loopback also fills both bytes with the lower byte).
*/

static void conduct_poll (CARD_ID card, BUS_DATA poll_response)
{
if ((di [card].cntl_register & CN_POLL) == CN_PPE)              /* if the card's poll response is enabled */
    poll_response |= PPR (DEV_ADDR (bus [card].dptr->flags));   /*   then add in the card's response */

tpprintf_bus (bus [card].dptr, BUS_PPOLL | poll_response, "Parallel poll\n");

if (poll_response) {                                    /* if a poll response is indicated */
    while (di [card].fifo_count != FIFO_SIZE)           /*   then fill the card FIFO with the responses */
        fifo_load (card, diag_access, poll_response);   /*     (hardware feature) */

    update_state (card);                                /* update the card state */
    }

return;
}


/* Accept a data byte from the bus.

   This routine is called when the card is in diagnostic mode and the other
   cross-connected card sources a data byte onto the bus.  It is responsible for
   loading the received data byte into the FIFO and updating the card state.  It
   returns TRUE if the card accepted the byte and FALSE if the card is not
   configured as a listener.

   On entry, the "card" parameter is the identifier of the receiving disc
   interface connected to the bus, and the "bus" parameter is the set of eight
   control and eight data lines representing the HP-IB.  If the indicated card
   is listening, it loads the byte into the FIFO and updates the card state to
   reflect the load.
*/

static t_bool di_bus_accept (CARD_ID card, BUS_CABLE signals)
{
if (di [card].cntl_register & CN_LSTN) {                /* if the interface is a listener */
    tpprintf_bus (bus [card].dptr, signals, "Data (card %u accepts)\n", card);

    fifo_load (card, bus_access, DATA (signals));       /* load the data byte into the FIFO */
    update_state (card);                                /*   and update the card state */

    return TRUE;                                        /* indicate that the byte was accepted */
    }

else                                                    /* otherwise the card is not listening */
    return FALSE;                                       /*   so it does not accept the byte */
}


/* Respond to the bus control lines.

   The indicated card is notified of the new control state on the bus.  The
   routine establishes the new bus state and updates the card state to reflect
   the change.

   If the card's control register has PPE set and CIC clear, the card's poll
   response is returned.
*/

static BUS_DATA di_bus_respond (CARD_ID card, BUS_CNTL new_cntl)
{
di [card].bus_cntl = new_cntl;                          /* update the bus control lines */
update_state (card);                                    /*   and the card state */

if ((di [card].cntl_register & CN_POLL) == CN_PPE)      /* if the card's poll response is enabled */
    return PPR (DEV_ADDR (bus [card].dptr->flags));     /*   then return the card's response */
else                                                    /* otherwise */
    return 0;                                           /*   the poll response is disabled */
}



/* Interface local utility routines */


/* Master reset the interface.

   This is the programmed card master reset, not the simulator reset routine.
   Master reset initializes a number of flip-flops and data paths on the card.
   The primary use, other than during a PRESET, is to clear the FIFO in
   preparation to changing the card from a listener to a talker or vice versa.
   This ensures that unneeded FIFO data is not transmitted inadvertently to the
   bus or to the CPU.  It is also used when changing the data mode from unpacked
   to packed to release the byte pointer flip-flops, which are held in the
   "lower byte" position during unpacked transfers.

   In hardware, a master reset:

    - clears the EDT, EOR, IRL, LBO, LBI, and IFC flip-flops

    - clears the Input Data Register

    - clears the FIFO

    - sets or clears the odd/even input and output byte pointer flip-flops,
      depending on whether the P (packed transfer) bit is set in the Control
      Word Register

   Master clear does not clear the IFC timer, so the IFC timer active and IFC
   flip-flop set status bits remain set if the timer is still active.


   Implementation notes:

    1. This routine is called during simulator startup before the device reset
       routine has been called (it is called from "cpu_reset" when that routine
       sends CRS and POPIO to all devices).  Therefore, we must validate the
       FIFO register pointer before clearing the queue index values.
*/

static void master_reset (CARD_ID card)
{
di [card].edt = CLEAR;                                  /* clear the EDT flip-flop */
di [card].eor = CLEAR;                                  /*   and the EOR flip-flop */

if (di [card].cntl_register & CN_PACK)                  /* if packed mode is set */
    di [card].ibp = di [card].obp = upper;              /*   then set the selectors to the upper byte */
else                                                    /* otherwise unpacked mode overrides */
    di [card].ibp = di [card].obp = lower;              /*   and sets the selectors to the lower byte */

di [card].status_register &= ~(ST_IRL | ST_LBO | ST_LBI);   /* clear the status flip-flops */

if (di [card].ifc_timer == 0.0)                         /* if the IFC timer is not active */
    di [card].status_register &= ~(ST_IFC | ST_IFCFF);  /*   then clear the IFC flip-flop */

di [card].input_data_register = 0;                      /* clear the input data register */

di [card].fifo_count = 0;                               /* clear the FIFO */

if (di [card].fifo_reg != NULL)                         /* if the FIFO queue pointer has been set up */
    di [card].fifo_reg->qptr =                          /*   then reset the FIFO */
      di [card].tag_reg->qptr = 0;                      /*     and TAG indices */

tpprintf (bus [card].dptr, TRACE_CSRW, "Master reset\n");
tpprintf (bus [card].dptr, TRACE_FIFO, "FIFO cleared\n");

return;
}


/* Update the interface state.

   In hardware, certain external operations cause automatic responses by the
   disc interface card.  For example, when the Input Data Register is unloaded
   by an LIx instruction, it is automatically reloaded with the next word from
   the FIFO.  Also, the card may be set to interrupt in response to the
   assertion of certain bus control lines.

   In simulation, this routine must be called whenever the FIFO, card control,
   or bus control state changes.  It determines whether:

    1. ...the next word from the FIFO should be unloaded into the IDR.  If the
       card is listening, and the IDR is empty, and the FIFO contains data, then
       a word is unloaded and stored in the IDR, and the Input Register Loaded
       status bit is set.

    2. ...the next word from the FIFO should be unloaded and sourced to the bus.
       If the card is talking (but not polling), and the listener is ready to
       accept data, and the last byte has not been sent, and the FIFO contains
       data, then a word is unloaded and sourced to the bus.  This occurs
       regardless of whether or not there are any listeners.

    3. ...an interface clear operation has completed.  If IFC is asserted, and
       the current simulation time is later than the IFC expiration time, then
       IFC is denied, and the timer is reset.

    4. ...the card should assert NRFD to prevent FIFO overflow.  If the card is
       listening, and the FIFO is full, or the last byte has been received, or a
       pause has been explicitly requested, then NRFD is asserted.

    5. ...the SRQ flip-flop should be set or cleared.  If the card is listening
       and the Input Data Register has been loaded, or the card is talking and
       the FIFO is not full, then SRQ is asserted to request a DCPC transfer.

    6. ...the Flag Buffer flip-flop should be set or cleared.  If the Input Data
       Register has been loaded or the Last Byte Out flip-flop is set and the
       corresponding Control Word Register IRL or LBO bits are set, or the End
       of Record flip-flop is set and the Input Data Register has been unloaded,
       or SRQ is asserted on the bus and the corresponding Control Word Register
       bit is set when the card is not the controller-in-charge, or REN or IFC
       is asserted on the bus and the corresponding Control Word Register bits
       are set when the card is not the system controller, then the Flag Buffer
       is set to request an interrupt.


   Implementation notes:

    1. The fifo_unload routine may set ST_LBO, so the test to set the Flag
       Buffer must be done after unloading.

    2. The gcc compiler (at least as of version 4.6.2) does not optimize
       repeated use of array-of-structures accesses.  Instead, it recalculates
       the index each time, even though the index is a constant within the
       function.  To avoid this performance penalty, we use a pointer to the
       selected DI_STATE structure.  Note that VC++ 2008 does perform this
       optimization.
*/

static void update_state (CARD_ID card)
{
DI_STATE * const di_card = &di [card];
HP_WORD   data;
FLIP_FLOP previous_srq;
BUS_CNTL  assert = BUS_NONE;
BUS_CNTL  deny = BUS_NONE;

if (di_card->cntl_register & CN_LSTN) {                 /* if the card is a listener */
    if (not (di_card->status_register & ST_IRL)         /*   then if the is IDR empty */
      && not FIFO_EMPTY) {                              /*     and data remains in the FIFO? */
        data = fifo_unload (card, cpu_access);          /*       then unload the FIFO */
        di_card->input_data_register = data;            /*         into the IDR */
        di_card->status_register |= ST_IRL;             /*           and set input register loaded status */
        }
    }

else if ((di_card->cntl_register                        /* otherwise if the card is a talker */
  & (CN_TALK | CN_PPE)) == CN_TALK)                     /*   and not polling */
    while (not FIFO_EMPTY                               /*     then while data remains in the FIFO */
      && not (di_card->bus_cntl & Bus_NRFD)             /*       and NRFD is denied */
      && not (di_card->status_register & ST_LBO)) {     /*       and the last byte has not been sent */
        data = fifo_unload (card, bus_access);          /*         then unload a FIFO byte */
        di_bus_source (card, di_card->bus_cntl | data); /*           and source it to the bus */
        }


if (di_card->bus_cntl & Bus_IFC) {                      /* if IFC is asserted */
    di_card->status_register |= ST_IFCFF;               /*   then set the IFC flip-flop */

    if (di_card->ifc_timer != 0.0                       /* if I asserted IFC and am still timing */
      && sim_gtime () > di_card->ifc_timer) {           /*   and the timeout has expired */
        di_card->ifc_timer = 0.0;                       /*     then clear the IFC timer */

        deny = Bus_IFC;                                 /* deny IFC on the bus */
        di_card->status_register &= ~ST_IFC;            /*   and clear IFC timer status */
        }
    }


if (di_card->cntl_register & CN_LSTN)                       /* if the card is listening */
    if (di_card->cntl_register & CN_NRFD                    /*   then if explicitly requested */
      || di_card->status_register & ST_LBI                  /*     or the last byte is in */
      || FIFO_FULL)                                         /*     or the FIFO is full */
        assert = Bus_NRFD ^ di_card->bus_cntl & Bus_NRFD;   /*       then assert NRFD if not already asserted */
    else                                                    /*   otherwise the card is ready for data */
        deny |= Bus_NRFD & di_card->bus_cntl;               /*     so deny NRFD if not already denied */

if (assert != deny)                                     /* if there was a change in bus state */
    di_bus_control (card, CONTROLLER, assert, deny);    /*   then update the bus control */


previous_srq = di_card->srq;                            /* save the current SRQ state */

if (di_card->cntl_register & CN_LSTN                    /* if the card is a listener */
  && di_card->status_register & ST_IRL                  /*   and the input register is loaded */
  || di_card->cntl_register & CN_TALK                   /* or the card is a talker */
  && not FIFO_FULL)                                     /*   and the FIFO is not full */
    di_card->srq = SET;                                 /*   then request a DCPC cycle */
else                                                    /* otherwise */
    di_card->srq = CLEAR;                               /*   DCPC service is not needed */


if (di_card->srq != previous_srq)                       /* if the SRQ state changed */
    tpprintf (bus [card].dptr, TRACE_CSRW, "SRQ %s\n",
              di_card->srq == SET ? "set" : "cleared");


if (di_card->status_register & ST_IRL                 /* if the input register is loaded */
   && di_card->cntl_register & CN_IRL                 /*   and notification is wanted */
  || di_card->status_register & ST_LBO                /* or the last byte is out */
   && di_card->cntl_register & CN_LBO                 /*   and notification is wanted */
  || di_card->eor == SET                              /* or the end of record was seen */
   && not (di_card->status_register & ST_IRL)         /*   and the input register was unloaded */
  || di_card->bus_cntl & Bus_SRQ                      /* or SRQ is asserted on the bus */
   && di_card->cntl_register & CN_SRQ                 /*   and notification is wanted */
   && di_card->cntl_register & CN_CIC                 /*   and the card is not the controller */
  || not SW8_SYSCTL                                   /* or the card is not the system controller */
   && di_card->bus_cntl & Bus_REN                     /*   and REN is asserted on the bus */
   && di_card->cntl_register & CN_REN                 /*   and notification is wanted */
  || not SW8_SYSCTL                                   /* or the card is not the system controller */
   && di_card->status_register & ST_IFCFF             /*   and the IFC flip-flop is set */
   && di_card->cntl_register & CN_IFC) {              /*   and notification is wanted */
    di_card->flag_buffer = SET;                       /*     then set the flag buffer */
    io_assert_ENF (bus [card].dptr->ctxt);            /*       and flag flip-flops and recalculate interrupts */
    }

else if (di_card->srq != previous_srq)                /* otherwise if SRQ changed state */
    io_assert_SIR (bus [card].dptr->ctxt);            /*   then recalculate the service request */

return;
}


/* Load a word or byte into the FIFO.

   A word or byte is loaded into the next available location in the FIFO.  The
   significance of the data parameter is indicated by the access mode as
   follows:

     - For CPU access, the parameter is a 16-bit value.

     - For bus access, the parameter is an 8-bit value in the lower byte and a
       zero in the upper byte.

     - For diagnostic access, the parameter is an 8-bit value in the lower byte
       that will be duplicated in the upper byte.

   For bus access, byte loading into the FIFO is controlled by the value of the
   Input Buffer Pointer (IBP) selector.

   In addition to the data words, the FIFO holds four tag bits that associate
   states or actions with each corresponding word.  They are:

     - Last Byte Received (while receiving, a line feed is received and the LF
       bit is set in the Control Word Register, or a byte with EOI asserted is
       received and the EOI bit is set).

     - End of Data Transfer (while transmitting, DCPC asserts the EDT
       backplane signal, or an unpacked-mode data word has the LBO bit set, or a
       packed-mode OTx is issued without an accompanying CLF).

     - ATN (the state of ATN on the bus if receiving, or the ATN bit in the
       unpacked data word if transmitting).

     - EOI (the state of EOI on the bus if receiving, or the EOI bit in the
       unpacked data word if transmitting).

   The tag is assembled into the upper word, the data is assembled into the
   lower word, and then the 32-bit value is stored in the next available FIFO
   location.

   If data is coming from the CPU, the 16-bit value is loaded into the next FIFO
   location, and the occupancy count is incremented.

   If the data is coming from the bus, and the input mode is unpacked, the 8-bit
   value is loaded into the lower byte of the next FIFO location, and the
   occupancy count is incremented.  In hardware, the upper FIFO is not clocked;
   in simulation, the upper byte is set to zero.  The IBP always points at the
   lower byte in unpacked mode.

   If the data is coming from the bus, and the input mode is packed, the 8-bit
   value is loaded into either the upper or lower byte of the next FIFO
   location, depending on the value of the IBP, and the IBP is toggled.  If the
   value was stored in the lower byte, the occupancy count is incremented.

   A special case occurs when the value is to be stored in the upper byte, and
   the LBR tag is set to indicate that this is the last byte to be received.  In
   this case, the value is stored in both bytes of the next FIFO location, and
   the occupancy counter is incremented.

   If data is coming from the diagnostic FIFO loopback, the 8-bit value in the
   lower byte is copied to the upper byte, the resulting 16-bit value is loaded
   into the next FIFO location, and the occupancy count is incremented.


   Implementation notes:

    1. The FIFO is implemented as circular queue to take advantage of REG_CIRC
       EXAMINE semantics.  REG->qptr is the index of the first word loaded into
       the FIFO and therefore represents the word present at the FIFO output.
       By specifying REG_CIRC, examining FIFO [0-n] and TAG [0-n] will always
       display the words in load order, regardless of the actual array index of
       the start of the list.  The number of words currently present in the FIFO
       is kept in fifo_count (0 = empty, 1-16 = number of words available).

       If fifo_count < FIFO_SIZE, "(REG->qptr + fifo_count) mod FIFO_SIZE" is
       the index of the new word location.  Loading stores the word there and
       then increments fifo_count.

    3. Because the load and unload routines need access to qptrs in the REG
       structures for the FIFO and TAG arrays, pointers to the REGs for each
       card are stored in the card state array during device reset.

    4. The FIFO and TAG register queue pointers are synchronized, so we need to
       check only one to determine the next location.

    5. The gcc compiler (at least as of version 4.6.2) does not optimize
       repeated use of array-of-structures accesses.  Instead, it recalculates
       the index each time, even though the index is a constant within the
       function.  To avoid this performance penalty, we use a pointer to the
       selected DI_STATE structure.  Note that VC++ 2008 does perform this
       optimization.
*/

static void fifo_load (CARD_ID card, FIFO_ACCESS access, HP_WORD data)
{
DI_STATE * const di_card = &di [card];
uint32 tag, index;
t_bool add_word = TRUE;

if (FIFO_FULL) {                                        /* if the FIFO is already full */
    tpprintf (bus [card].dptr, TRACE_FIFO,
              access == bus_access ? "Attempted load of data %02XH into full FIFO\n"
                                   : "Attempted load of data %06o into full FIFO\n",
              data);
    return;                                             /*   then return with the load ignored */
    }

if (di_card->cntl_register & CN_LSTN) {                     /* if the card is receiving */
    tag = TO_TAG (di_card->bus_cntl & (Bus_ATN | Bus_EOI)); /*   then set the tag from the bus signals */

    if ((di_card->cntl_register & CN_EOI                /* if EOI detection is enabled */
       && di_card->bus_cntl & Bus_EOI)                  /*   and data was tagged with EOI */
      || (di_card->cntl_register & CN_LF                /* or LF detection is enabled */
       && LOWER_BYTE (data) == LF)) {                   /*   and the byte is a line feed */
        tag = tag | TAG_LBR;                            /*     then tag as the last byte received */
        di_card->status_register |= ST_LBI;             /*       and set the last byte in status */
        }

    else                                                /* otherwise neither termination condition was seen */
        di_card->status_register &= ~ST_LBI;            /*   so clear the last byte in status */
    }

else                                                    /* otherwise the card is transmitting */
    tag = TO_TAG (data & (DATA_ATN | DATA_EOI));        /*   so set the tag from the data */

if (di_card->edt == SET)                                /* if this is the end of the data transfer */
    tag |= TAG_EDT;                                     /*   then add the EDT tag */

index = (di_card->fifo_reg->qptr                        /* calculate the index */
          + di_card->fifo_count) % FIFO_SIZE;           /*   of the next available location */

if (access == bus_access) {                             /* if this is a bus access */
    if (di_card->ibp == upper) {                        /*   then if the last was upper byte access */
        di_card->ibp = lower;                           /*     then set the lower byte as next */

        if (tag & TAG_LBR)                              /* if this is the last byte */
            di_card->fifo [index] =                     /*   then copy to both bytes to the FIFO */
              tag | TO_WORD (data, data);               /*     and store with the tag */

        else {                                          /* otherwise more bytes are expected */
            di_card->fifo [index] =                     /*   so position this byte */
              tag | TO_WORD (data, 0);                  /*     and store it with the tag */

            add_word = FALSE;                           /* wait for the second byte before adding */
            }
        }

    else                                                /* otherwise the lower byte was last */
        if (di_card->cntl_register & CN_PACK) {         /*   so if the card is in packed mode */
            di_card->ibp = upper;                       /*     then set the upper byte as next */

            di_card->fifo [index] =                     /* merge the data and tag values */
              tag | di_card->fifo [index] | TO_WORD (0, data);
            }

        else                                            /* otherwise the card is in unpacked mode */
            di_card->fifo [index] =                     /*   so position this byte */
              tag | TO_WORD (0, data);                  /*     and store with the tag */
    }

else if (access == cpu_access)                          /* otherwise if this is a cpu access */
    di_card->fifo [index] = tag | data;                 /*   then store the tag and full word in the FIFO */

else {                                                  /* otherwise it must be a diagnostic access */
    data = TO_WORD (data, data);                        /*   so copy the lower byte to the upper byte */
    di_card->fifo [index] = tag | data;                 /*     and store the tag and full word in the FIFO */
    }

if (add_word)                                           /* if we added a word to the FIFO */
    di_card->fifo_count++;                              /*   then increment the count of words stored */

tpprintf (bus [card].dptr, TRACE_FIFO,
          access == bus_access ? "Data %02XH tag %s loaded into FIFO [%u] count %u\n"
                               : "Data %06o tag %s loaded into FIFO [%u] count %u\n",
          data, fmt_bitset (tag, tag_format), index, di_card->fifo_count);
return;
}


/* Unload a word or byte from the FIFO.

   A word or byte is unloaded from the first location in the FIFO.  The
   significance of the returned value is indicated by the access mode as
   follows:

     - For CPU access, a 16-bit value is unloaded and returned.

     - For bus access, an 8-bit value is unloaded and returned.

     - For diagnostic access, an 16-bit value is unloaded, and the lower byte
       is returned.

   For bus access, byte unloading from the FIFO is controlled by the value of
   the Output Buffer Pointer (OBP) selector.

   In addition to the data words, the FIFO holds four tag bits that associate
   states or actions with each corresponding word.  They are:

     - Last Byte Received (sets the End of Record flip-flop when the last byte
       received is loaded into the Input Data Register).

     - End of Data Transfer (sets the LBO bit in the Status Word Register when
       the last byte is unloaded from the FIFO).

     - ATN (in unpacked mode, sets the ATN bit in the returned data word if
       listening, or controls the bus ATN line if talking; in packed mode, the
       tag is ignored).

     - EOI (in unpacked mode, sets the EOI bit in the returned data word if
       listening, or asserts the bus EOI line if talking; in packed mode, the
       tag is ignored).

   If the FIFO is not empty, the first entry is obtained and split into tag and
   data words.  The LBR tag value is loaded into the EOR flip-flop if the CPU is
   accessing.  The EDT tag sets Last Byte Out status if the last byte is being
   unloaded.

   If the data is going to the CPU, the 16-bit packed data value is returned as
   is, or the lower byte of the unpacked value is merged with the tags for ATN
   and EOI and returned.  The occupancy count is decremented to unload the FIFO
   entry.

   If the data is going to the bus, and the input mode is unpacked, the 8-bit
   value is returned in the lower byte, and the occupancy count is decremented.
   In hardware, the upper FIFO is not clocked; in simulation, the upper byte is
   ignored.  The OBP always points at the lower byte in unpacked mode.

   If the data is going to the bus, and the input mode is packed, the 8-bit
   value is unloaded from either the upper or lower byte of the data word,
   depending on the value of the OBP, and returned in the lower byte.  The OBP
   value is toggled.  If the value was obtained from the lower byte, the
   occupancy count is decremented to unload the FIFO.  Otherwise, the count is
   not altered, so that the lower-byte access will be from the same FIFO entry.

   If data is going to the diagnostic FIFO loopback, the lower byte of the
   16-bit value is returned; the upper byte of the returned value is zero.


   Implementation notes:

    1. ATN and EOI tag handling is complex.  If the card is listening in the
       unpacked mode, the ATN tag substitutes for bit 8 of the data word, and
       the EOI tag substitutes for bit 9.  In the packed mode, bits 8 and 9 are
       as stored in the FIFO (they are upper-byte data bits).

       If the card is talking in the unpacked mode, the ATN tag asserts or
       denies ATN on the bus if the card is the CIC, and the EOI tag asserts or
       denies EOI on the bus.  In the packed mode, the ATN bit in the Control
       Word Register asserts or denies ATN on the bus if the card is the CIC,
       and the EOI bit asserts EOI on the bus if the last byte of the entry
       tagged with EDT has been unloaded from the FIFO (which sets LBO status)
       or denies EOI otherwise.

    2. In hardware, the EOR flip-flop is clocked with the Input Data Register.
       Therefore, when the card is listening, EOR is set not when the last byte
       is unloaded from the FIFO, but rather when that byte is loaded into the
       IDR.  These two actions occur together when the IDR is empty.

       However, during diagnostic access, data unloaded from the FIFO is
       reloaded, and the IDR is never clocked.  As the T and L bits must be set
       with DIAG in the Control Word Register to enable the loopback path, the
       LBR tag will be entered into the FIFO if EOI or LF detection is enabled,
       but the EOR flip-flop will not be set when that word falls through to be
       unloaded.

       In simulation, EOR is set whenever the LBR tag is unloaded from the FIFO
       during CPU access, as a CPU unload is always followed by an IDR store.

    3. If fifo_count > 0, REG->qptr is the index of the word to remove.  Removal
       gets the word and then increments the FIFO and TAG qptrs (mod FIFO_SIZE)
       and decrements fifo_count.

    4. The FIFO and TAG register queue pointers are synchronized, so we need to
       check only one to determine the first location.

    5. The gcc compiler (at least as of version 4.6.2) does not optimize
       repeated use of array-of-structures accesses.  Instead, it recalculates
       the index each time, even though the index is a constant within the
       function.  To avoid this performance penalty, we use a pointer to the
       selected DI_STATE structure.  Note that VC++ 2008 does perform this
       optimization.
*/

static HP_WORD fifo_unload (CARD_ID card, FIFO_ACCESS access)
{
DI_STATE * const di_card = &di [card];
uint32 data, tag, index;
t_bool remove_word = TRUE;

if (FIFO_EMPTY) {                                       /* if the FIFO is already empty */
    tpprintf (bus [card].dptr, TRACE_FIFO, "Attempted unload from empty FIFO\n");
    return 0;                                           /*   then return with dummy data */
    }

index = di_card->fifo_reg->qptr;                        /* save the index for the trace */
data = di_card->fifo [index];                           /*   and get the tag and data from the FIFO */

tag = data & TAG_MASK;                                  /* mask the tag to just the tag bits */
data = data & D16_MASK;                                 /*   and the data to just the data bits */

if (tag & TAG_EDT                                       /* if this is the end of a data transfer */
  && (di_card->obp == lower                             /*   and the lower byte is next */
  || di_card->cntl_register & CN_ODD))                  /*   or we are sending an odd number of bytes */
    di_card->status_register |= ST_LBO;                 /*     then set the last byte out status */


if (access == cpu_access) {                             /* if this is a cpu access */
    if (not (di_card->cntl_register & CN_PACK))         /*   then if the card is in unpacked mode */
        data = data & ~(DATA_ATN | DATA_EOI)            /*     then substitute the ATN/EOI tag values */
          | TO_BUS (tag & (TAG_ATN | TAG_EOI));         /*       into the data word */

    di_card->eor = D_FF (tag & TAG_LBR);                /* set the end-of-record flip-flop for the last byte */
    }

else if (access == bus_access)                          /* otherwise if this is a bus access */
    if (di_card->obp == upper) {                        /*   then if this is the upper byte */
        di_card->obp = lower;                           /*     then set the lower byte as next */

        data = UPPER_BYTE (data);                       /* mask and position the upper byte in the data word */
        remove_word = FALSE;                            /*   but do not unload the FIFO until the next byte */
        }

    else {                                              /* otherwise this is the lower byte */
        data = LOWER_BYTE (data);                       /*   so mask and position it in the data word */

        if (di_card->cntl_register & CN_PACK)           /* if the card is in the packed mode */
            di_card->obp = upper;                       /*   then set the upper byte as next */
        }

else                                                    /* otherwise this must be a diagnostic access */
    data = LOWER_BYTE (data);                           /*   so access is to the lower byte only */


if (remove_word) {                                      /* if the word was removed from the FIFO */
    di_card->fifo_reg->qptr =                           /*   then update the FIFO queue pointer */
      di_card->tag_reg->qptr =                          /*     and the TAG pointer */
      (di_card->fifo_reg->qptr + 1) % FIFO_SIZE;        /*       and wrap around as needed */

    di_card->fifo_count--;                              /* decrement the FIFO count */
    }

tpprintf (bus [card].dptr, TRACE_FIFO,
          access == bus_access ? "Data %02XH tag %s unloaded from FIFO [%u] count %u\n"
                               : "Data %06o tag %s unloaded from FIFO [%u] count %u\n",
          data, fmt_bitset (tag, tag_format), index, di_card->fifo_count);

if (di_card->cntl_register & CN_TALK)                   /* if the card is talking */
    if (di_card->cntl_register & CN_PACK)               /*   then if it in the packed mode */
        if (di_card->status_register & ST_LBO           /*     then if this is the last byte out */
          && di_card->cntl_register & CN_EOI)           /*       and is EOI control is enabled */
            di_card->bus_cntl |= Bus_EOI;               /*         then assert EOI on the bus */
        else                                            /*     otherwise */
            di_card->bus_cntl &= ~Bus_EOI;              /*       deny EOI on the bus */

    else {                                              /* otherwise the card is in the unpacked mode */
        if (di_card->cntl_register & CN_CIC)            /*   so if the card the controller in charge */
            di_card->bus_cntl =                         /*     then assert or deny the ATN bus line */
              di_card->bus_cntl & ~Bus_ATN              /*       from the ATN tag value */
                | TO_BUS (tag & TAG_ATN);

        di_card->bus_cntl =                             /* assert or deny the EOI bus line */
          di_card->bus_cntl & ~Bus_EOI                  /*   from the EOI tag value */
            | TO_BUS (tag & TAG_EOI);
        }

return (HP_WORD) data;                                  /* return the unloaded data value */
}
