/* hp2100_mpt.c: HP 12584A Sixteen-Channel Teleprinter Multiplexer Interface simulator

   Copyright (c) 2022, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from the author.

   MPT          12584A-01 Teleprinter Multiplexer Interface
   MPD          12584A-02 Data Set Multiplexer Interface

   17-Mar-22    JDB     First release
   20-Feb-22    JDB     First successful 2000A logon
   19-Feb-22    JDB     Passes the 24202A diagnostic
   17-Feb-22    JDB     Created the MPT and MPD devices

   References:
     - 12584A, 12584A-01, 12584A-02 Teleprinter Multiplexor Computer Interface Kits
         Operating and Service Manual (12584-9001, May 1969)


   The 12584A is a 16-channel terminal multiplexer designed explicitly for
   teleprinters (e.g., Teletypes).  It supports direct-connected terminals as
   well as modems at a fixed speed of 110 baud.  It is the original terminal
   multiplexer for the HP 2000A, B, and C Time-Shared BASIC systems.  It is
   replaced by the later 12920A 16-channel asynchronous multiplexer in the
   2000C' (C-prime, a.k.a., High-Speed C), E, F, and Access systems.

   The multiplexer interface consists of a 16-bit output register, a 16-bit
   input buffer, a fixed-frequency oscillator that runs at 880 Hz, and
   associated control and flag circuitry.  The output register is connected
   through TTL-to-RS-232 level shifters to the serial Received Data (BB) lines
   of up to 16 terminals.  The input buffer contains 16 RS-232-to-TTL level
   shifters connected to the terminals' serial Transmitted Data (BA) lines.  The
   oscillator is connected to the flag buffer flip-flop to generate an interrupt
   every 1.136 milliseconds, providing eight interrupts per 110 baud serial bit
   time.  All serial-to-parallel and parallel-to-serial conversions are
   performed in software by the CPU, with the interface serving as a pair of
   16-bit level shifters and a periodic interrupt source.

   A single terminal multiplexer card may be cabled to up to 16 direct-connected
   terminals.  The multiplexer provides very limited dataset control: all
   sixteen Data Terminal Ready (CD) terminal lines are connected together and
   raised or lowered with the state of the control flip-flop.

   To permit remote operation with modems, a second multiplexer is used to
   control the individual DTR lines and to sense the Ring Indicator (CE) and
   Data Carrier Detect (CF) signals.  The dataset multiplexer is identical to
   the terminal multiplexer, except that the oscillator is not connected to the
   flag circuitry, and so the interface does not interrupt periodically.

   This simulation provides one terminal multiplexer (MPT device) and one
   dataset multiplexer (MTD device).  Terminal emulators may be connected via
   Telnet or host serial communication ports.  Attaching a listening port to the
   MPT device allows incoming Telnet connections to be made to multiplexer
   ports.  Individual ports may also be attached to specific host serial ports.
   Dropping DTR, either collectively when the MTD device is disabled, or
   individually via the associated MTD output line, causes Telnet sessions to be
   disconnected and serial port DTR lines to be denied.

   The simulation provides a DIAGNOSTIC mode that simulates replacement of the
   terminal and dataset cables with a loopback connector.  This permits the
   multiplexer diagnostic to be run.  The simulation also provides REALTIME and
   FASTTIME modes.  The former must be used with the diagnostic to pass the
   timing tests.  The latter improves communication rates somewhat.  However,
   because all of the serial conversions are done in software, there is a limit
   to the speed improvement possible before CPU saturation occurs.  The default
   FASTTIME setting improves the rate about three-to-one, resulting in about a
   5% decrease in available CPU time.


   Implementation notes:

     1. The interface card buffers the flag buffer flip-flop with a "flag
        storage" flip-flop.  Its purpose is to record an oscillator tick that
        occurs before the interrupt service routine has completed and reset the
        flag flip-flop.  The simulator implements this feature.

     2. Correction factors must be applied to the oscillator timing to pass the
        diagnostic timing tests.  The factors adjust the default 1000 E-Series
        timing to match that of the 2100/2116/2115 and 2114 CPUs.

     3. The 2100 diagnostic requires configuration of the oscillator speed to
        one of five rates, corresponding to 110, 150, 220, 300, and 330 baud.
        The 12584A hardware oscillator provides a trim adjustment but otherwise
        has no provision for changing the rate.
*/



#include <ctype.h>

#include "sim_defs.h"
#include "sim_tmxr.h"

#include "hp2100_defs.h"
#include "hp2100_io.h"



/* Program limits */

#define CARD_COUNT          2                   /* the count of cards supported */

#define MPT_PORTS           16                  /* the number of visible units */
#define MPT_CNTLS           2                   /* the number of control units */


/* Device property constant declarations */

#define PORT_DELAY_2100     (int32) mS (1.136 * 1.011)  /* 880 Hz * 2100/2116 correction factor */
#define PORT_DELAY_2114     (int32) mS (1.136 * 0.774)  /* 880 Hz * 2114 correction factor */
#define PORT_DELAY_FAST     (int32) mS (0.379)          /* 2640 Hz (330 baud) */


/* Device flags */

#define DEV_DIAG_SHIFT      (DEV_V_UF + 0)      /* the diagnostic connector is installed */
#define DEV_REALTIME_SHIFT  (DEV_V_UF + 1)      /* the timing mode is realistic */

#define DEV_DIAG            (1u << DEV_DIAG_SHIFT)
#define DEV_REALTIME        (1u << DEV_REALTIME_SHIFT)


/* Unit flags */

#define UNIT_CAPS_SHIFT     (TTUF_V_UF + 0)     /* the terminal CAPS LOCK key is down */

#define UNIT_CAPSLOCK       (1u << UNIT_CAPS_SHIFT)


/* Unit references */

#define CARD                u3                          /* the card ID unit field */

#define mpt_port            mpt_unit [MPT_PORTS + 0]    /* the terminal port unit */
#define mpt_poll            mpt_unit [MPT_PORTS + 1]    /* the terminal poll unit */

#define mpd_port            mpd_unit [0]                /* the dataset port unit */


static const BITSET_NAME port_names [] = {      /* port bit assignment names */
    "0",                                        /*   bit  0 */
    "1",                                        /*   bit  1 */
    "2",                                        /*   bit  2 */
    "3",                                        /*   bit  3 */
    "4",                                        /*   bit  4 */
    "5",                                        /*   bit  5 */
    "6",                                        /*   bit  6 */
    "7",                                        /*   bit  7 */
    "8",                                        /*   bit  8 */
    "9",                                        /*   bit  9 */
    "10",                                       /*   bit 10 */
    "11",                                       /*   bit 11 */
    "12",                                       /*   bit 12 */
    "13",                                       /*   bit 13 */
    "14",                                       /*   bit 14 */
    "15"                                        /*   bit 15 */
    };

static const BITSET_FORMAT port_format =        /* names, offset, direction, alternates, bar */
    { FMT_INIT (port_names, 0, lsb_first, no_alt, no_bar) };


typedef enum {                                  /* multiplexer card ID numbers */
    mpt,                                        /*   the teleprinter multiplexer card index */
    mpd                                         /*   the dataset multiplexer card index */
    } CARD_INDEX;


/* Interface state */

typedef struct {
    HP_WORD      output_data;                   /* output data register */
    HP_WORD      input_data;                    /* input data buffer */
    FLIP_FLOP    control;                       /* control flip-flop */
    FLIP_FLOP    flag;                          /* flag flip-flop */
    FLIP_FLOP    flag_buffer;                   /* flag buffer flip-flop */
    FLIP_FLOP    flag_storage;                  /* flag storage flip-flop */
    } CARD_STATE;

static CARD_STATE state [CARD_COUNT];           /* per-card state */

static int32 fast_time = PORT_DELAY_FAST;       /* fast port service delay */


/* Multiplexer per-line state */

static uint32 tx_data   [MPT_PORTS];            /* the current transmit character */
static uint32 rx_data   [MPT_PORTS];            /* the current receive character */
static uint32 tx_cntr   [MPT_PORTS];            /* the transmit sample counter */
static uint32 rx_cntr   [MPT_PORTS];            /* the receive sample counter */
static uint32 rx_buffer [MPT_PORTS];            /* the receive input buffer */


/* Interface local SCP support routines */

static IO_INTERFACE   mpt_interface;
static IO_INITIALIZER mpt_initializer;


/* Interface local SCP support routines */

static t_stat port_service (UNIT   *uptr);
static t_stat poll_service (UNIT   *uptr);
static t_stat mpt_reset    (DEVICE *dptr);
static t_stat mpd_reset    (DEVICE *dptr);
static t_stat mpt_attach   (UNIT   *uptr, char *cptr);
static t_stat mpt_detach   (UNIT   *uptr);

static t_stat set_mode    (UNIT *uptr, int32 value, char  *cptr, void *desc);
static t_stat show_mode   (FILE *st,   UNIT  *uptr, int32 value, void *desc);
static t_stat show_status (FILE *st,   UNIT  *uptr, int32 value, void *desc);


/* Interface local support routines */

static t_stat  connect_port     (UNIT *uptr, int32 value, char *cptr, void *desc);
static HP_WORD dataset_status   (void);
static void    control_dataset  (HP_WORD new_state);
static void    control_terminal (TMCKT   control);

/* Interface SCP data structures */

DEVICE mpt_dev;                                 /* incomplete device structure */
DEVICE mpd_dev;                                 /* incomplete device structure */


/* Teleprinter multiplexer library descriptors */

static TMLN mpt_ldsc [MPT_PORTS] = {            /* the line descriptors */
    { 0 }
    };

static TMXR mpt_desc = {                        /* the multiplexer descriptor */
    MPT_PORTS,                                  /*   the number of terminal lines */
    0,                                          /*   the listening port (reserved) */
    0,                                          /*   the master socket  (reserved) */
    mpt_ldsc,                                   /*   the line descriptor array */
    NULL,                                       /*   the line connection order */
    &mpt_dev                                    /*   the multiplexer device */
    };


/* Device information block */

static DIB mpt_dib = {                          /* the teleprinter multiplexer DIB */
    mpt_interface,                              /*   the device's I/O interface function pointer */
    mpt_initializer,                            /*   the execution initialization function pointer */
    NULL,                                       /*   the power state function pointer */
    MPT,                                        /*   the device's select code (02-77) */
    mpt,                                        /*   the card index */
    "12584A Teleprinter Multiplexer Interface", /*   the card description */
    NULL                                        /*   the ROM description */
    };

static DIB mpd_dib = {                          /* the dataset multiplexer DIB */
    mpt_interface,                              /*   the device's I/O interface function pointer */
    mpt_initializer,                            /*   the execution initialization function pointer */
    NULL,                                       /*   the power state function pointer */
    MPD,                                        /*   the device's select code (02-77) */
    mpd,                                        /*   the card index */
    "12584A Dataset Multiplexer Interface",     /*   the card description */
    NULL                                        /*   the ROM description */
    };


/* Unit lists */

#define UNIT_FLAGS          (TT_MODE_7P | UNIT_CAPSLOCK)
#define PORT_FLAGS          (UNIT_DIS)
#define POLL_FLAGS          (UNIT_ATTABLE | UNIT_DIS)

static UNIT mpt_unit [] = {
/*           Event Routine  Unit Flags  Capacity    Delay     Card */
/*           -------------  ----------  --------  ----------  ---- */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  0 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  1 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  2 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  3 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  4 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  5 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  6 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  7 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  8 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line  9 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line 10 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line 11 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line 12 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line 13 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line 14 */
    { UDATA (NULL,          UNIT_FLAGS,    0),        0,      mpt  },   /* terminal I/O line 15 */

    { UDATA (&port_service, PORT_FLAGS,    0),        0,      mpt  },   /* terminal port unit */
    { UDATA (&poll_service, POLL_FLAGS,    0),    POLL_FIRST, mpt  }    /* terminal poll unit */
    };

static UNIT mpd_unit [] = {
/*           Event Routine  Unit Flags  Capacity    Delay     Card */
/*           -------------  ----------  --------  ----------  ---- */
    { UDATA (&port_service,     0,         0),        0,      mpd  }    /* dataset port unit */
    };


/* Register lists */

static REG mpt_reg [] = {
/*    Macro   Name    Location                   Radix  Width  Offset     Depth      Flags  */
/*    ------  ------  -------------------------  -----  -----  ------  -----------  ------- */
    { ORDATA (INPUT,  state [mpt].input_data,            16)                                },
    { ORDATA (OUTPUT, state [mpt].output_data,           16)                                },

    { BRDATA (RDATA,  rx_data,                     8,     8,           MPT_PORTS),  REG_A   },
    { BRDATA (RCNTR,  rx_cntr,                    10,    16,           MPT_PORTS),  PV_LEFT },
    { BRDATA (RBUF,   rx_buffer,                   8,     8,           MPT_PORTS),  REG_A   },
    { BRDATA (TDATA,  tx_data,                     8,     8,           MPT_PORTS),  REG_A   },
    { BRDATA (TCNTR,  tx_cntr,                    10,    16,           MPT_PORTS),  PV_LEFT },

    { FLDATA (CTL,    state [mpt].control,                       0)                         },
    { FLDATA (FLG,    state [mpt].flag,                          0)                         },
    { FLDATA (FBF,    state [mpt].flag_buffer,                   0)                         },
    { FLDATA (FST,    state [mpt].flag_storage,                  0)                         },
    { DRDATA (TIME,   fast_time,                         24),                       PV_LEFT },

      DIB_REGS (mpt_dib),

    { NULL }
    };

static REG mpd_reg [] = {
/*    Macro   Name    Location                   Width  Offset   Flags  */
/*    ------  ------  -------------------------  -----  ------  ------- */
    { ORDATA (INPUT,  state [mpd].input_data,     16)                   },
    { ORDATA (OUTPUT, state [mpd].output_data,    16)                   },
    { FLDATA (CTL,    state [mpd].control,                0)            },
    { FLDATA (FLG,    state [mpd].flag,                   0)            },
    { FLDATA (FBF,    state [mpd].flag_buffer,            0)            },
    { FLDATA (FST,    state [mpd].flag_storage,           0)            },

      DIB_REGS (mpd_dib),

    { NULL }
    };


/* Modifier lists */

typedef enum {                                  /* Device modes */
    Fast_Time,                                  /*   use optimized timing */
    Real_Time,                                  /*   use realistic timing */
    Terminal,                                   /*   use the terminal connector and cable */
    Dataset,                                    /*   use the dataset connector and cable */
    Diagnostic                                  /*   use the diagnostic connector */
    } DEVICE_MODES;

static MTAB mpt_mod [] = {
/*    Mask Value     Match Value    Print String      Match String  Validation  Display  Descriptor */
/*    -------------  -------------  ----------------  ------------  ----------  -------  ---------- */
    { UNIT_CAPSLOCK, UNIT_CAPSLOCK, "CAPS LOCK down", "CAPSLOCK",   NULL,       NULL,    NULL       },
    { UNIT_CAPSLOCK, 0,             "CAPS LOCK up",   "NOCAPSLOCK", NULL,       NULL,    NULL       },

    { TT_MODE,       TT_MODE_UC,    "UC",             "UC",         NULL,       NULL,    NULL       },
    { TT_MODE,       TT_MODE_7B,    "7b",             "7B",         NULL,       NULL,    NULL       },
    { TT_MODE,       TT_MODE_8B,    "8b",             "8B",         NULL,       NULL,    NULL       },
    { TT_MODE,       TT_MODE_7P,    "7p",             "7P",         NULL,       NULL,    NULL       },

/*    Entry Flags          Value       Print String   Match String  Validation       Display           Descriptor         */
/*    -------------------  ----------  -------------  ------------  ---------------  ----------------  ------------------ */
    { MTAB_XDV,            Fast_Time,  NULL,          "FASTTIME",   &set_mode,       NULL,             (void *) &mpt_dev  },
    { MTAB_XDV,            Real_Time,  NULL,          "REALTIME",   &set_mode,       NULL,             (void *) &mpt_dev  },
    { MTAB_XDV,            Terminal,   NULL,          "TERMINAL",   &set_mode,       NULL,             (void *) &mpt_dev  },
    { MTAB_XDV,            Diagnostic, NULL,          "DIAGNOSTIC", &set_mode,       NULL,             (void *) &mpt_dev  },
    { MTAB_XDV,            0,          "MODES",       NULL,         NULL,            &show_mode,       (void *) &mpt_dev  },

    { MTAB_XUN | MTAB_NC,  0,          "LOG",         "LOG",        &tmxr_set_log,   &tmxr_show_log,   (void *) &mpt_desc },
    { MTAB_XUN | MTAB_NC,  0,          NULL,          "NOLOG",      &tmxr_set_nolog, NULL,             (void *) &mpt_desc },
    { MTAB_XUN,            0,          NULL,          "DISCONNECT", &tmxr_dscln,     NULL,             (void *) &mpt_desc },

    { MTAB_XDV,            0,          "",            NULL,         NULL,            &show_status,     (void *) &mpt_desc },
    { MTAB_XDV | MTAB_NMO, 1,          "CONNECTIONS", NULL,         NULL,            &tmxr_show_cstat, (void *) &mpt_desc },
    { MTAB_XDV | MTAB_NMO, 0,          "STATISTICS",  NULL,         NULL,            &tmxr_show_cstat, (void *) &mpt_desc },
    { MTAB_XDV,            0,          NULL,          "CONNECT",    &connect_port,   NULL,             NULL               },

    { MTAB_XDV,             1u,        "SC",          "SC",         &hp_set_dib,     &hp_show_dib,     (void *) &mpt_dib  },
    { MTAB_XDV | MTAB_NMO, ~1u,        "DEVNO",       "DEVNO",      &hp_set_dib,     &hp_show_dib,     (void *) &mpt_dib  },

    { 0 }
    };

static MTAB mpd_mod [] = {
/*    Entry Flags          Value       Print String   Match String  Validation   Display       Descriptor        */
/*    -------------------  ----------  -------------  ------------  -----------  ------------  ----------------- */
    { MTAB_XDV,            Dataset,    NULL,          "DATASET",    &set_mode,   NULL,         (void *) &mpd_dev },
    { MTAB_XDV,            Diagnostic, NULL,          "DIAGNOSTIC", &set_mode,   NULL,         (void *) &mpd_dev },
    { MTAB_XDV,            0,          "MODES",       NULL,         NULL,        &show_mode,   (void *) &mpd_dev },

    { MTAB_XDV,             1u,        "SC",          "SC",         &hp_set_dib, &hp_show_dib, (void *) &mpd_dib },
    { MTAB_XDV | MTAB_NMO, ~1u,        "DEVNO",       "DEVNO",      &hp_set_dib, &hp_show_dib, (void *) &mpd_dib },

    { 0 }
    };


/* Debugging trace lists */

static DEBTAB mpt_deb [] = {                    /* teleprinter multiplexer trace list */
    { "CSRW",  TRACE_CSRW  },                   /*   trace interface control, status, read, and write actions */
    { "PSERV", TRACE_PSERV },                   /*   trace periodic unit service scheduling calls and entries */
    { "XFER",  TRACE_XFER  },                   /*   trace data transmissions and receptions */
    { "IOBUS", TRACE_IOBUS },                   /*   trace I/O bus signals and data words received and returned */
    { NULL,    0           }
    };

static DEBTAB mpd_deb [] = {                    /* dataset multiplexer trace list */
    { "CSRW",  TRACE_CSRW  },                   /*   trace interface control, status, read, and write actions */
    { "IOBUS", TRACE_IOBUS },                   /*   trace I/O bus signals and data words received and returned */
    { NULL,    0           }
    };


/* Device descriptors */

DEVICE mpt_dev = {
    "MPT",                                      /* device name */
    mpt_unit,                                   /* unit array */
    mpt_reg,                                    /* register array */
    mpt_mod,                                    /* modifier array */
    MPT_PORTS + MPT_CNTLS,                      /* number of units */
    10,                                         /* address radix */
    31,                                         /* address width */
    1,                                          /* address increment */
    8,                                          /* data radix */
    16,                                         /* data width */
    &tmxr_ex,                                   /* examine routine */
    &tmxr_dep,                                  /* deposit routine */
    &mpt_reset,                                 /* reset routine */
    NULL,                                       /* boot routine */
    &mpt_attach,                                /* attach routine */
    &mpt_detach,                                /* detach routine */
    &mpt_dib,                                   /* device information block pointer */
    DEV_DISABLE | DEV_DIS | DEV_DEBUG,          /* device flags */
    0,                                          /* debug control flags */
    mpt_deb,                                    /* debug flag name array */
    NULL,                                       /* memory size change routine */
    NULL                                        /* logical device name */
    };

DEVICE mpd_dev = {
    "MPD",                                      /* device name */
    mpd_unit,                                   /* unit array */
    mpd_reg,                                    /* register array */
    mpd_mod,                                    /* modifier array */
    1,                                          /* number of units */
    10,                                         /* address radix */
    31,                                         /* address width */
    1,                                          /* address increment */
    8,                                          /* data radix */
    16,                                         /* data width */
    NULL,                                       /* examine routine */
    NULL,                                       /* deposit routine */
    &mpd_reset,                                 /* reset routine */
    NULL,                                       /* boot routine */
    NULL,                                       /* attach routine */
    NULL,                                       /* detach routine */
    &mpd_dib,                                   /* device information block pointer */
    DEV_DISABLE | DEV_DIS | DEV_DEBUG,          /* device flags */
    0,                                          /* debug control flags */
    mpd_deb,                                    /* debug flag name array */
    NULL,                                       /* memory size change routine */
    NULL                                        /* logical device name */
    };

static DEVICE *dptrs [CARD_COUNT] = {           /* device pointer array, indexed by CARD_INDEX */
    &mpt_dev,
    &mpd_dev
    };



/* Multiplexer interface.

   The multiplexer interface is installed on the I/O bus and receives I/O
   commands from the CPU.  In simulation, the asserted signals on the bus are
   represented as bits in the inbound_signals set.  Each signal is processed
   sequentially in ascending numerical order.  The outbound signals and optional
   data value are returned after inbound signal processing is complete.

   In DIAGNOSTIC mode, the interface behaves as though a loopback connector
   is installed.  The connector also connects the oscillator output to the flag
   circuitry to provide periodic interrupts.  This is significant when the
   dataset interface is in diagnostic mode, as the dataset cable used in the
   DATASET mode does not connect the oscillator.

   The interface contains a second flag buffer, called the "flag storage"
   flip-flop.  Its purpose is to record an oscillator tick if the interrupt
   caused by the prior tick has not been serviced.  When the tick occurs, the
   flag storage flip-flop sets.  If the flag flip-flop is clear, indicating that
   the interrupt service routine has completed, the flag buffer flip-flop will
   be set when SIR asserts.  This, in turn, will set the flag flip-flop and
   clear the flag storage flip-flop when ENF asserts.  If the flag remains set,
   indicating that the prior interrupt is still being serviced, the flag storage
   flip-flop will also remain set.  Once the flag is cleared at the end of the
   interrupt service routine, the flag buffer will be set as above.


   Implementation notes:

    1. The teleprinter multiplexer input data word is assembled in the port
       service routine and is available when an IOI signal is asserted.  The
       output data word is disassembled in the port service routine and only
       needs to be recorded for an IOO signal.  For the dataset multiplexer, the
       input word is assembled by the "dataset_status" routine, and the output
       word is disassembled by the "control_dataset" routine that are called by
       the IOI and IOO handlers, respectively.

    2. Setting and clearing the teleprinter multiplexer's control flip-flop
       performs different actions, depending on whether or not the dataset
       multiplexer is present.  This is due to differing hardware configurations
       of the multiplexer terminal connector panel.  If the dataset card is not
       present, then the teleprinter control flip-flop also controls the sixteen
       DTR lines collectively.  If the card is present, then the teleprinter
       control signal has no effect, as the DTR lines are controlled
       individually by the dataset card.  Presence or absence of the dataset
       card is determined by whether the MPD device is enabled or disabled.
*/

static SIGNALS_VALUE mpt_interface (const DIB *dibptr, INBOUND_SET inbound_signals, HP_WORD inbound_value)
{
const CARD_INDEX card = (CARD_INDEX) dibptr->card_index;    /* the card selector */
DEVICE * const   dptr = dptrs [card];                       /* the associated device pointer */
INBOUND_SIGNAL   signal;
INBOUND_SET      working_set = inbound_signals;
SIGNALS_VALUE    outbound    = { ioNONE, 0 };
t_bool           irq_enabled = FALSE;

while (working_set) {                                   /* while signals remain */
    signal = IONEXTSIG (working_set);                   /*   isolate the next signal */

    switch (signal) {                                   /* dispatch the I/O signal */

        case ioCLF:                                     /* Clear Flag flip-flop */
            state [card].flag_buffer = CLEAR;           /* clear the flag buffer */
            state [card].flag        = CLEAR;           /*   and flag flip-flops */

            if (state [card].flag_storage == SET) {     /* if the storage flip-flop is holding a request */
                state [card].flag_buffer = SET;         /*   then set the flag buffer */
                working_set |= ioENF;                   /*     and ensure the flag is enabled */
                }
            break;


        case ioSTF:                                     /* Set Flag flip-flop */
            state [card].flag_buffer = SET;             /* set the flag buffer flip-flop */
            break;


        case ioENF:                                     /* Enable Flag */
            if (state [card].flag_buffer == SET) {      /* if the flag buffer flip-flop is set */
                state [card].flag = SET;                /*   then set the flag flip-flop */
                state [card].flag_storage = CLEAR;      /*     and clear the flag storage flip-flop */
                }
            break;


        case ioSFC:                                     /* Skip if Flag is Clear */
            if (state [card].flag == CLEAR)             /* if the flag flip-flop is clear */
                outbound.signals |= ioSKF;              /*   then assert the Skip on Flag signal */
            break;


        case ioSFS:                                     /* Skip if Flag is Set */
            if (state [card].flag == SET)               /* if the flag flip-flop is set */
                outbound.signals |= ioSKF;              /*   then assert the Skip on Flag signal */
            break;


        case ioIOI:                                                 /* I/O data input */
            if (dptr->flags & DEV_DIAG)                             /* if the card is in diagnostic mode */
                state [card].input_data = state [card].output_data; /*   then loop the data back */

            else if (card == mpd)                               /* otherwise if this is the dataset card */
                state [card].input_data = dataset_status ();    /*   then return the line status */

            outbound.value = state [card].input_data;   /* set the outbound data value from the input buffer */
            break;


        case ioIOO:                                             /* I/O data output */
            if (card == mpd && not (dptr->flags & DEV_DIAG))    /* if the dataset card is not in diagnostic mode */
                control_dataset (inbound_value);                /*   then set the control lines */

            state [card].output_data = inbound_value;   /* store the inbound data value in the output buffer */
            break;


        case ioPOPIO:                                   /* Power-On Preset to I/O */
            state [card].flag_buffer = SET;             /* set the flag buffer flip-flop */
            state [card].output_data = D16_UMAX;        /*   and the output register to all ones */
            break;


        case ioCRS:                                     /* Control Reset */
        case ioCLC:                                     /* Clear Control flip-flop */
            if (mpd_dev.flags & DEV_DIS                 /* if the dataset card is not present */
              && not (mpt_dev.flags & DEV_DIAG)         /*   and the terminal card is not in diagnostic mode */
              && state [card].control == SET)           /*     and the control flip-flop is currently set */
                control_terminal (No_Signals);          /*       then clearing it drops DTR on all terminals */

            state [card].control = CLEAR;               /* clear the control flip-flop */
            break;


        case ioSTC:                                     /* Set Control flip-flop */
            if (mpd_dev.flags & DEV_DIS                 /* if the dataset card is not present */
              && not (mpt_dev.flags & DEV_DIAG)         /*   and the terminal card is not in diagnostic mode */
              && state [card].control == CLEAR)         /*     and the control flip-flop is currently clear */
                control_terminal (DTR_Control);         /*       then setting it raises DTR on all terminals */

            state [card].control = SET;                 /* set the control flip-flop */
            break;


        case ioSIR:                                         /* Set Interrupt Request */
            if (state [card].control & state [card].flag)   /* if the control and flag flip-flops are set */
                outbound.signals |= cnVALID;                /*   then deny PRL */
            else                                            /* otherwise */
                outbound.signals |= cnPRL | cnVALID;        /*   conditionally assert PRL */

            if (state [card].control & state [card].flag    /* if the control and flag */
              & state [card].flag_buffer)                   /*   and flag buffer flip-flops are set */
                outbound.signals |= cnIRQ;                  /*     then conditionally assert IRQ */

            if (state [card].flag == SET)               /* if the flag flip-flop is set */
                outbound.signals |= ioSRQ;              /*   then assert SRQ */
            break;


        case ioIAK:                                     /* Interrupt Acknowledge */
            state [card].flag_buffer = CLEAR;           /* clear the flag buffer flip-flop */
            break;


        case ioIEN:                                     /* Interrupt Enable */
            irq_enabled = TRUE;                         /* permit IRQ to be asserted */
            break;


        case ioPRH:                                         /* Priority High */
            if (irq_enabled && outbound.signals & cnIRQ)    /* if IRQ is enabled and conditionally asserted */
                outbound.signals |= ioIRQ | ioFLG;          /*   then assert IRQ and FLG */

            if (not irq_enabled || outbound.signals & cnPRL)    /* if IRQ is disabled or PRL is conditionally asserted */
                outbound.signals |= ioPRL;                      /*   then assert it unconditionally */
            break;


        case ioEDT:                                     /* not used by this interface */
        case ioPON:                                     /* not used by this interface */
            break;
        }

    IOCLEARSIG (working_set, signal);                   /* remove the current signal from the set */
    }                                                   /*   and continue until all signals are processed */

return outbound;                                        /* return the outbound signals and value */
}


/* Multiplexer initializer.

   This routine is called from the instruction execution prelude to set up the
   internal state of the multiplexer prior to instruction execution.

   The routine sets up the teleprinter port timing delay to be used to simulate
   ticks of the internal oscillator.  The delay is dependent on the REALTIME vs.
   FASTTIME setting, as well as the CPU model.

   The dataset card oscillator is only used in DIAGNOSTIC mode, when the
   loopback connector is in place.  The delay depends on the CPU model only;
   there is no FASTTIME setting for the oscillator, as the diagnostic tests the
   expected tick rate and will fail if the rate is too fast or too slow.
*/

static t_bool mpt_initializer (DEVICE *dptr, DIB *dibptr)
{
const CARD_INDEX card = (CARD_INDEX) dibptr->card_index;    /* the card selector */
int32            delay;

if (cpu_configuration & CPU_2114)           /* if the CPU is a 2114 */
    delay = PORT_DELAY_2114;                /*   then use the 2114 timing */
else                                        /* otherwise */
    delay = PORT_DELAY_2100;                /*   use the 2116/2100 timing */

if (card == mpt)                            /* if this is the terminal card */
    if (dptr->flags & DEV_REALTIME)         /*   then if real-time mode is enabled */
        mpt_port.wait = delay;              /*     then use the realistic time delay */
    else                                    /*   otherwise */
        mpt_port.wait = fast_time;          /*     use the optimized time delay */

else                                        /* otherwise this is the dataset card */
    mpd_port.wait = delay;                  /*   whose real-time delay is used only for the diagnostic */

return TRUE;
}


/* Service the teleprinter ports.

   The port service is called to receive and send characters from and to the
   multiplexer ports.  The state of each serial line is reflected in the data
   bit of the card's input buffer and output register corresponding to the port
   number. Because serial-to-parallel and parallel-to-serial conversions are
   handled in software by the target operating system's interrupt service
   routine, we must perform the opposite operation here to convert the character
   data handled by the simulator to serial data exchanged with the CPU.

   In hardware, the interface interrupts eight times faster than the serial data
   rate, so that each serial bit persists on its corresponding parallel data
   word bit for eight interrupts.  In simulation, this service routine is called
   eight times for each bit transferred.  The target OS assumes one start bit
   and two stop bits are present, resulting in eleven serial bits or 88 entries
   per character.

   Each service entry in TERMINAL mode assembles a 16-bit input data word from
   the computed serial state of the sixteen multiplexer ports.  Per-port
   counters are employed to control when the serial bits change.  If a port
   counter is zero, the line is idle; otherwise, the value counts down the
   remaining number of service entries needed to complete reception.  Every
   eight counts, the current character is shifted right one bit.  The LSB of the
   character is set into the corresponding input data word bit.

   Characters are received from Telnet or serial connections by the separate
   input poll service routine and are placed in the per-port receive buffers.  A
   waiting character will have either the SCPE_BREAK or TMXR_VALID bits set; if
   no character is waiting, the buffer value will be zero.  Reception of either
   a BREAK or character while the line is idle sets the counter to the number of
   service entries needed to serialize the character, including the start and
   stop bits.  Subsequent service entries are handled as above.

   After all port receptions are handled, the transmit section scans the ports
   for character transmissions in progress.  A separate transmission counter is
   used to assemble a complete output character.  Because the bits are received
   LSB first, each serial bit is copied into bit 7 (the sign bit of the byte),
   and the byte is right-shifted once every eight service entries.  When the
   counter has been decremented to zero, the character is set to the port's
   Telnet or serial connection.

   When the output port is idle, transmission assembly is initiated by the
   presence of a start bit on the serial line.  This sets up the transmission
   counter, which proceeds to count down on every subsequent service entry while
   handling the serial-to-parallel conversion as above.

   As part of reception or transmission, character conversions are performed as
   directed by the port's CAPSLOCK and TTY mode flags.  An output stall sets the
   transmission counter to 1, so that output will be retried with the next
   service entry.

   Once both reception and transmission are processed, the flag storage
   flip-flop is set.  If the flag is clear, the flag buffer is also set, and ENF
   is asserted to inform the CPU of the pending interrupt.  The service routine
   is then rescheduled for the next entry.

   In DIAGNOSTIC mode, the service routine simply interrupts without performing
   any reception or transmission operations.


   Implementation notes:

    1. To be detected, HP 2000 Time-Shared BASIC requires that a serial break
       must hold the output line at 0 for at least 100 interrupts.  We
       arbitrarily choose 112 interrupts (14 bit times) to ensure detection.

    2. The transmit counter is set to 72 interrupts (9 bit times) so that
       transmission completes and the character is sent to the Telnet or serial
       connection before the two stop bits are output by the OS.  This slightly
       improves service performance.

    3. TSB outputs characters with the parity bit set (always on).  This is
       masked by the "7B" and "7P" output modes but will be apparent if "8B"
       mode is selected.

    4. The terminal multiplexer library "tmxr_putc_ln" routine returns
       SCPE_STALL if it is called when the transmit buffer is full.  When the
       last character is added to the buffer, the routine returns SCPE_OK but
       also changes the "xmte" field of the terminal multiplexer line (TMLN)
       structure from 1 to 0 to indicate that further calls will be rejected.
       The "xmte" value is set back to 1 when the transmit buffer empties.

       This presents two approaches to handling buffer overflows: either call
       "tmxr_putc_ln" unconditionally and test for SCPE_STALL on return, or call
       "tmxr_putc_ln" only if "xmte" is 1.  The former approach adds a new
       character to the transmit buffer as soon as space is available, while the
       latter adds a new character only when the buffer has completely emptied.
       With either approach, transmission must be rescheduled to allow the
       buffer to drain.

       It would seem that the former approach is more attractive, as it would
       allow the simulated I/O operation to complete more quickly.  However,
       there are two mitigating factors.  First, the library attempts to write
       the entire transmit buffer in one host system call, so there is usually
       no time difference between freeing one buffer character and freeing the
       entire buffer (barring host system buffer congestion).  Second, the
       routine increments a "character dropped" counter when returning
       SCPE_STALL status.  However, the characters actually would not be lost,
       as the SCPE_STALL return would schedule retransmission when buffer space
       is available.  This would lead to erroneous reporting in the SHOW <unit>
       STATISTICS command.

       Therefore, we adopt the latter approach and reschedule transmission if
       the "xmte" field is 0.  Note that the "tmxr_poll_tx" routine still must
       be called in this case, as it is responsible for transmitting the buffer
       contents and therefore freeing space in the buffer.

    5. The "tmxr_putc_ln" library routine returns SCPE_LOST if the line is not
       connected.  We ignore this error so that an OS may output a logoff
       message even when the terminal has disconnected.  This permits the
       simulation to continue while ignoring the output.
*/

static t_stat port_service (UNIT *uptr)
{
const CARD_INDEX card = (CARD_INDEX) uptr->CARD;    /* the card index */
DEVICE * const   dptr = dptrs [card];               /* the associated device pointer */
const uint32     frame_bits = 03000;                /* one start bit and two stop bits (shifted) */
int32            cvtd_data;
uint32           port;
t_stat           result;
t_bool           char_sent = FALSE;

tpprintf (dptr, TRACE_PSERV, "Port delay %d service entered\n",
          uptr->wait);

if (not (dptr->flags & DEV_DIAG)) {                     /* if the card is not in diagnostic mode */

/* Reception section */

    state [card].input_data = D16_UMAX;                 /* start with all line bits = 1 */

    for (port = 0; port < MPT_PORTS; port++)                /* loop through the set of ports */
        if (rx_cntr [port] > 0) {                           /* if the port is currently receiving a character */
            if ((rx_data [port] & 1) == 0)                  /*   then if the current bit is 0 */
                state [card].input_data &= ~(1u << port);   /*     then clear corresponding line bit */

            rx_cntr [port]--;                           /* decrement the sample counter */

            if (rx_cntr [port] % 8 == 0)                /* if this ends sampling of the current bit */
                rx_data [port] >>= 1;                   /*   then move the next bit into place */
            }

        else if (rx_buffer [port] != 0)                 /* otherwise if a new character is waiting */
            if (rx_buffer [port] & SCPE_BREAK) {        /*   then if the BREAK key was pushed */
                rx_buffer [port] = 0;                   /*     then clear the buffered character */
                rx_data   [port] = 0;                   /*       and the current character */

                rx_cntr [port] = 111;                   /* set the sample counter to 14 bit times - 1 */

                tpprintf (dptr, TRACE_XFER, "Port %d break detected\n",
                          port);
                }

            else {                                      /* otherwise a normal character is waiting */
                tpprintf (dptr, TRACE_XFER, "Port %d character %s received\n",
                          port, fmt_char (LOWER_BYTE (rx_buffer [port])));

                rx_data [port] = LOWER_BYTE (rx_buffer [port]); /* get the new character */

                if (mpt_unit [port].flags & UNIT_CAPSLOCK)              /* if the CAPS LOCK key is down */
                    rx_data [port] = (uint32) toupper (rx_data [port]); /*   then upshift the character */

                rx_data [port] = rx_data [port] << 1 | frame_bits;  /* align and merge the framing bits */

                rx_buffer [port] = 0;                   /* clear the buffer */

                state [card].input_data &= ~(1u << port);   /* set the start bit on the line */
                rx_cntr [port] = 87;                        /*   and set the counter to 11 bit times - 1 */
                }

/* Transmission section */

    for (port = 0; port < MPT_PORTS; port++)            /* loop through the set of ports */
        if (tx_cntr [port] > 0) {                       /* if the port is currently transmitting a character */
            if (tx_cntr [port] % 8 == 0) {              /*   then if this ends sampling of the current bit */
                tx_data [port] >>= 1;                   /*     then make room for the next bit */

                if (state [card].output_data & 1u << port)  /* if the next bit is 1 */
                    tx_data [port] |= D8_SIGN;              /*   then set the corresponding data bit */
                }

            tx_cntr [port]--;                           /* decrement the sample counter */

            if (tx_cntr [port] == 0) {                      /* if all of the bits have been accumulated */
                cvtd_data = sim_tt_outcvt (tx_data [port],  /*   then convert the character for the output mode */
                                           TT_GET_MODE (mpt_unit [port].flags));

                if (cvtd_data == 0)                     /* if the character was filtered out */
                    tpprintf (dptr, TRACE_XFER, "Port %d character %s discarded by output filter\n",
                              port, fmt_char (tx_data [port] & D8_SMAX));

                else if (mpt_ldsc [port].xmte == 0) {   /* otherwise it's printable but transmission is stalled */
                    tx_cntr [port] = 1;                 /*   so retry at the next service */
                    char_sent = TRUE;                   /*     and try to flush the line buffer */

                    tpprintf (dptr, TRACE_XFER, "Port %d character %s transmission stalled for full buffer\n",
                              port, fmt_char (cvtd_data & D8_SMAX));
                    }

                else {                                          /* otherwise */
                    result = tmxr_putc_ln (&mpt_ldsc [port],    /*   add the character to the output buffer */
                                           cvtd_data);

                    if (result == SCPE_OK) {            /* if the addition succeeded */
                        char_sent = TRUE;               /*   then mark the line buffer to be flushed */

                        tpprintf (dptr, TRACE_XFER, "Port %d character %s%s transmitted\n",
                                  port, fmt_char (cvtd_data & D8_SMAX),
                                  (cvtd_data & D8_SIGN ? " + odd parity" : ""));
                        }

                    else if (result == SCPE_LOST)       /* otherwise if the connection was lost */
                        tpprintf (dptr, TRACE_XFER, "Port %d character %s discarded by connection loss\n",
                                  port, fmt_char (cvtd_data & D8_SMAX));
                    }
                }
            }

        else if ((state [card].output_data & 1u << port) == 0) {    /* otherwise if a start bit is present */
            tx_data [port] = 0;                                     /*  then initialize the new character */
            tx_cntr [port] = 71;                                    /*    and set the counter to 9 bit times - 1 */
            }

    if (char_sent)                                      /* if characters were written */
        tmxr_poll_tx (&mpt_desc);                       /*   then flush the output buffers */
    }

state [card].flag_storage = SET;                        /* set the flag storage flip-flop */

if (state [card].flag == CLEAR)                         /* if the flag flip-flop is clear */
    state [card].flag_buffer = SET;                     /*   then set the flag buffer */

io_assert_ENF (dptrs [card]->ctxt);                     /* enable the flag */

sim_activate (uptr, uptr->wait);                        /* reschedule the next service */

tpprintf (dptr, TRACE_PSERV, "Port delay %d service rescheduled\n",
          uptr->wait);

return SCPE_OK;                                         /* return success */
}


/* Service the multiplexer connection poll.

   This service routine is used to poll for connections and incoming characters.
   It is started when the listening socket or a serial line is attached and is
   stopped when the socket and all lines are detached.

   Incoming characters are added to the port reception buffers only when they
   are free.  The port service will begin reception of a buffered character
   automatically when the port is or becomes idle.

   The poll is synchronized with the master keyboard poll timer to permit
   idling, although this is not currently possible when running HP 2000
   Time-Shared BASIC.
*/

static t_stat poll_service (UNIT *uptr)
{
uint32 port;
int32  new_port;

tprintf (mpt_dev, TRACE_PSERV, "Poll delay %d service entered\n",
         uptr->wait);

new_port = tmxr_poll_conn (&mpt_desc);                  /* check for a new connection */

if (new_port >= 0)                                      /* if a new connection was established */
    connect_port (NULL, new_port, NULL, NULL);          /*   then set up the multiplexer port */

tmxr_poll_rx (&mpt_desc);                               /* poll for incoming data */

for (port = 0; port < MPT_PORTS; port++)                    /* loop through the set of ports */
    if (rx_buffer [port] == 0)                              /* if the buffer is free */
        rx_buffer [port] = tmxr_getc_ln (&mpt_ldsc [port]); /*   then get a new character if present */

if (uptr->wait == POLL_FIRST)                           /* if this is the first poll */
    uptr->wait = hp_sync_poll (INITIAL);                /*   then synchronize with the master poll unit */
else                                                    /* otherwise */
    uptr->wait = hp_sync_poll (SERVICE);                /*   maintain poll unit synchronization */

sim_activate (uptr, uptr->wait);                        /* continue polling */

tprintf (mpt_dev, TRACE_PSERV, "Poll delay %d service rescheduled\n",
         uptr->wait);

return SCPE_OK;                                         /* return success */
}


/* Teleprinter multiplexer reset routine.

   This routine is called for a RESET or RESET MPT command.  It ensures
   activation of the port and poll service routines if the listening port is
   attached.  If the multiplexer is in DIAGNOSTIC mode, it ensures that the port
   service is running.  If neither condition prevails, then the port and poll
   services are stopped.

   If a RESET -P (power-on reset) is done, the default FASTTIME setting is
   restored.
*/

static t_stat mpt_reset (DEVICE *dptr)
{
if (sim_switches & SWMASK ('P'))                        /* if this is the power-on reset */
    fast_time = PORT_DELAY_FAST;                        /*   then restore the original optimized timing */

if (mpt_poll.flags & UNIT_ATT) {                        /* if the listening port is attached */
    mpt_poll.wait = POLL_FIRST;                         /*   then set the initial poll delay */
    sim_activate (&mpt_poll, mpt_poll.wait);            /*     and schedule the poll service */
    sim_activate (&mpt_port, mpt_port.wait);            /*       and the port service */
    }

else if (dptr->flags & DEV_DIAG)                        /* otherwise if the card is in diagnostic mode */
    sim_activate (&mpt_port, mpt_port.wait);            /*   then schedule the port service */

else {                                                  /* otherwise the card is idle */
    sim_cancel (&mpt_poll);                             /*   so stop the poll */
    sim_cancel (&mpt_port);                             /*     and port services */
    }

return SCPE_OK;                                         /* return success */
}


/* Dataset multiplexer reset routine.

   This routine is called for a RESET or RESET MPD command.  It ensures
   activation of the port service routine if the multiplexer is in DIAGNOSTIC
   mode.  Otherwise, the port service is stopped, as the dataset connector does
   not connect the oscillator to the flag circuitry.
*/

static t_stat mpd_reset (DEVICE *dptr)
{
if (dptr->flags & DEV_DIAG)                             /* if the card is in diagnostic mode */
    sim_activate (&mpd_port, mpd_port.wait);            /*   then schedule the port service */
else                                                    /* otherwise */
    sim_cancel (&mpd_port);                             /*   the port service does not run */

return SCPE_OK;                                         /* return success */
}


/* Attach the teleprinter multiplexer or a multiplexer port.

   We are called by the ATTACH MPT <netport> command to attach the multiplexer
   to the network listening port indicated by <netport> and by ATTACH MPT<n>
   <serport> to attach line <n> to serial port <serport>.  If the attach
   succeeds, the port and poll services are scheduled to service the
   connections.


   Implementation notes:

    1. If we are being called as part of RESTORE processing, we may see a
       request to attach the poll unit (unit 17).  This will occur if the
       listening port was attached when the SAVE was done.  If the user attempts
       to attach unit 16 (port unit) or 17 directly, the command will be
       rejected by SCP because the units are disabled.

    2. If the poll unit was attached during a SAVE, it will be enabled as part
       of RESTORE processing.  We always unilaterally disable the poll unit to
       ensure that it remains hidden.

    3. ATTACH MPT will pass a pointer unit 0.  This is because SCP treats ATTACH
       MPT as equivalent to ATTACH MPT0.  The "tmxr_attach_unit" routine
       differentiates these cases by examining the "sim_ref_type" global to see
       if a device was referenced and, if so, applying the request to the poll
       unit.
*/

static t_stat mpt_attach (UNIT *uptr, char *cptr)
{
t_stat status;

status = tmxr_attach_unit (&mpt_desc, &mpt_poll, uptr, cptr);   /* attach the line */

if (status == SCPE_OK) {                                /* if the attach succeeded */
    mpt_poll.wait = POLL_FIRST;                         /*   then set the initial poll delay */
    sim_activate (&mpt_poll, mpt_poll.wait);            /*     and schedule the poll service */
    sim_activate (&mpt_port, mpt_port.wait);            /*       and the port service */

    tprintf (mpt_dev, TRACE_PSERV, "Poll delay %d service scheduled\n",
             mpt_poll.wait);

    mpt_poll.flags |= UNIT_DIS;                         /* ensure that the poll unit remains disabled */
    }

return status;                                          /* return the result of the attach */
}


/* Detach the teleprinter multiplexer or a multiplexer line.

   We are called by the DETACH MPT command to detach the listening port and all
   Telnet sessions and by the DETACH MPT<n> to detach a serial port from line
   <n>.  We will also be called by DETACH ALL, RESTORE, and during simulator
   shutdown.  For DETACH ALL and RESTORE, we must not fail the call, or
   processing of other units will cease.

   Detaching the listening port will disconnect all active Telnet connections
   but will not disturb any lines connected to serial ports.  Therefore, we look
   for newly freed lines and clean them up.  If all lines are now free, then we
   terminate the port and poll events.


   Implementation notes:

    1. During simulator shutdown, we will be called for units 0-16 (detach_all
       in scp.c calls the detach routines of all units that do NOT have
       UNIT_ATTABLE), as well as for unit 17 if it is attached.

    2. Because DETACH MPT will pass unit 0, the "tmxr_detach_unit" routine
       examines the "sim_ref_type" global to see if a device was referenced and,
       if so, applies the request to the poll unit.  The routine will reject an
       attempt to detach unit 16 (port unit) or unit 17 (poll unit) directly.
*/

static t_stat mpt_detach (UNIT *uptr)
{
t_stat status;
uint32 line;

status = tmxr_detach_unit (&mpt_desc, &mpt_poll, uptr); /* detach the line */

if (status == SCPE_OK) {                                /* if the detach succeeded */
    for (line = 0; line < MPT_PORTS; line++)            /*   then look for Telnet connections */
        if (tmxr_line_free (&mpt_ldsc [line]))          /* if the line is now free */
            mpt_ldsc [line].rcve = 0;                   /*   then disable reception */

    if (tmxr_mux_free (&mpt_desc)) {                    /* if all lines are now free */
        sim_cancel (&mpt_poll);                         /*   then stop polling */
        sim_cancel (&mpt_port);

        tprintf (mpt_dev, TRACE_PSERV, "Poll service stopped\n");
        }
    }

return status;                                          /* return the status of the detach */
}


/* Set the device modes.

   This validation routine is entered with the "value" parameter set to one of
   the DEVICE_MODES values.  The device flag implied by the value is set or
   cleared.  The unit, character, and descriptor pointers are not used.

   If the mode is being changed from terminal to diagnostic, the listening port
   and all serial connections are detached.  The port service is then
   rescheduled to provide interrupts from the oscillator for the diagnostic.

   If the mode is being changed from dataset to diagnostic, the port service is
   scheduled, as the diagnostic connector loops the oscillator output to the
   flag circuitry.  The normal dataset connector leaves the oscillator output
   uncommitted.

   If the mode is being changed from diagnostic to terminal or dataset, the poll
   routine is canceled because no network connections will exist.


   Implementation notes:

    1. The Fast_Time and Real_Time values are used only for the teleprinter
       multiplexer card.

    2. The real-time service delay depends on the CPU model, so it must be set
       by the device initializer routine.
*/

static t_stat set_mode (UNIT *uptr, int32 value, char *cptr, void *desc)
{
DEVICE * const dptr = (DEVICE *) desc;
uint32 port;

switch ((DEVICE_MODES) value) {                         /* dispatch the mode to set */

    case Fast_Time:                                     /* entering optimized timing mode */
        dptr->flags &= ~DEV_REALTIME;                   /*   so clear the real-time flag */
        break;


    case Real_Time:                                     /* entering realistic timing mode */
        dptr->flags |= DEV_REALTIME;                    /*   so set the real-time flag */
        break;


    case Terminal:                                      /* entering terminal mode */
        dptr->flags &= ~DEV_DIAG;                       /*   so clear the diagnostic flag */
        sim_cancel (&mpt_port);                         /*     and stop the port service */
        break;


    case Dataset:                                       /* entering dataset mode */
        dptr->flags &= ~DEV_DIAG;                       /*   so clear the diagnostic flag */
        sim_cancel (&mpd_port);                         /*     and stop the port service */
        break;


    case Diagnostic:                                    /* entering diagnostic mode */
        dptr->flags |= DEV_DIAG;                        /*   so set the diagnostic flag */

        if (uptr->CARD == mpd)                          /* if this is the dataset card */
            sim_activate (&mpd_port, mpd_port.wait);    /*   then connect the oscillator to the flag */

        else {                                          /* otherwise it is the terminal card */
            mpt_detach (&mpt_poll);                     /*   so detach the listening port and Telnet connections */

            for (port = 0; port < MPT_PORTS; port++) {          /* detach any existing */
                tmxr_detach_line (&mpt_desc, &mpt_unit [port]); /*   serial connections */
                mpt_ldsc [port].rcve = 0;                       /*     and disable port reception */
                }

            sim_activate (&mpt_port, mpt_port.wait);    /* connect the oscillator to the flag storage flip-flop */
            }
        break;
    }

return SCPE_OK;                                         /* mode changes always succeed */
}


/* Show the device modes.

   This display routine is called to show the device modes for the multiplexers.
   The output stream is passed in the "st" parameter, a pointer to the device is
   passed in the "desc" parameter, and the "uptr" and "value" parameters are
   ignored.  The connection mode and the timing mode for the teleprinter
   multiplexer are printed.
*/

static t_stat show_mode (FILE *st, UNIT *uptr, int32 value, void *desc)
{
DEVICE * const dptr = (DEVICE *) desc;

if (uptr->CARD == mpt)                                  /* if this is the terminal card */
    fprintf (st, "%s timing, %s mode",                  /*   then print the timing and connection modes */
             (dptr->flags & DEV_REALTIME ? "realistic"  : "fast"),
             (dptr->flags & DEV_DIAG     ? "diagnostic" : "terminal"));

else                                                    /* otherwise */
    fprintf (st, "%s mode",                             /*   print just the connection mode */
             (dptr->flags & DEV_DIAG ? "diagnostic" : "dataset"));

return SCPE_OK;                                         /* return success */
}


/* Show the teleprinter multiplexer status */

static t_stat show_status (FILE *st, UNIT *uptr, int32 value, void *desc)
{
if (mpt_poll.flags & UNIT_ATT)                                  /* if the listening port is attached */
    fprintf (st, "attached to port %s, ", mpt_poll.filename);   /*   then report the port number */
else                                                            /* otherwise */
    fprintf (st, "not attached, ");                             /*   report that the port is not attached */

tmxr_show_summ (st, uptr, value, desc);                 /* report the connection summary */

return SCPE_OK;                                         /* return success */
}


/* Set up a new connection.

   This validation routine is called to initialize a new teleprinter multiplexer
   port connection, optionally waiting for the port to connect.  On entry,
   "uptr" points at a UNIT structure if a connection wait is desired or is NULL
   if the connection has already been established.  For the former case, the
   actual value of "uptr" is irrelevant, and the "port" parameter is ignored.
   For the latter, "port" indicates the number of the multiplexer port to be
   initialized.  The "cptr" and "desc" parameters are not used.

   A connection wait is typically requested with a SET <dev> CONNECT command
   whose MTAB entry specifies this validation routine.  The routine waits until
   a connection to the listening socket is established.  The user may abort the
   wait by entering the SCP interrupt character, which defaults to CTRL+E.
   Connections that are established via the poll service routine pass NULL for
   the "uptr" parameter to indicate that the wait is not needed.
*/

t_stat connect_port (UNIT *uptr, int32 port, char *cptr, void *desc)
{
t_stat status;

if (uptr != NULL)                                       /* if a connection wait is requested */
    status = tmxr_connect (&mpt_desc, &port);           /*   then wait for the incoming request */
else                                                    /* otherwise the "port" parameter */
    status = SCPE_OK;                                   /*   already contains the new port number */

if (status == SCPE_OK) {                                /* if connection is established */
    mpt_ldsc [port].rcve = 1;                           /*   then enable the port to receive */

    tprintf (mpt_dev, TRACE_CSRW, "Port %d connection established\n", port);
    }

return status;                                          /* return the connection status */
}


/* Get the current status of the port datasets.

   This routine is called to assemble the current dataset status word.  Each bit
   in the word corresponds to the state of a port's Data Carrier Detect or Ring
   Indicator line.  The current state of these lines is obtained, and the bit
   corresponding to the port number is asserted (cleared) if either signal is
   active.
*/

static HP_WORD dataset_status (void)
{
uint32  port;
TMCKT   status;
HP_WORD prior_state, new_state, asserted, denied;

prior_state = state [mpd].input_data;                   /* save the prior state for comparison */
new_state   = D16_UMAX;                                 /*   and start the new state with inactive lines */

for (port = 0; port < MPT_PORTS; port++) {              /* loop through the set of ports */
    status = tmxr_line_status (&mpt_ldsc [port]);       /* get the port's line status */

    if (status & (DCD_Status | RI_Status))              /* if either DCD or RI are asserted */
        new_state &= ~(1u << port);                     /*   then mark the port as active */
    }

if (prior_state != new_state) {                         /* if any port changed its status */
    asserted =  prior_state & ~new_state;               /*   then get the new line assertions */
    denied   = ~prior_state &  new_state;               /*     and the new line denials */

    if (asserted)                                               /* if there are any new assertions */
        tprintf (mpd_dev, TRACE_CSRW, "Port %s asserted DCD\n", /*   then report them */
                 fmt_bitset (asserted, port_format));

    if (denied)                                                 /* if there are any new denials */
        tprintf (mpd_dev, TRACE_CSRW, "Port %s denied DCD\n",   /*   then report them */
                 fmt_bitset (denied, port_format));
    }

return new_state;                                       /* return the current port status set */
}


/* Control the port datasets individually.

   This routine is called to set the Data Terminal Ready line of each device
   connected to a multiplexer port.  Each bit in the "new_state" parameter
   corresponds to the state of a port's DTR line.  The line is asserted if the
   bit corresponding to the port number is active (cleared).

   To improve performance, the "tmxr_control_line" routine is called only when
   a port's DTR state changes.
*/

static void control_dataset (HP_WORD new_state)
{
uint32  port;
HP_WORD prior_state, asserted, denied;

prior_state = state [mpd].output_data;                  /* save the prior state for comparison */

if (prior_state != new_state) {                         /* if control of any port changed */
    asserted =  prior_state & ~new_state;               /*   then get the new line assertions */
    denied   = ~prior_state &  new_state;               /*     and the new line denials */

    if (asserted)
        tprintf (mpd_dev, TRACE_CSRW, "Port %s DTR asserted\n", /* if there are any new assertions */
                 fmt_bitset (asserted, port_format));           /*   then report them */

    if (denied)                                                 /* if there are any new denials */
        tprintf (mpd_dev, TRACE_CSRW, "Port %s DTR denied\n",   /*   then report them */
                 fmt_bitset (denied, port_format));

    for (port = 0; port < MPT_PORTS; port++)            /* loop through the set of ports */
        if (asserted & 1u << port)                      /* if control is asserted for the port */
            tmxr_control_line (&mpt_ldsc [port],        /*   then raise DTR on the port */
                               DTR_Control);

        else if (denied & 1u << port)                   /* otherwise if control is denied for the port */
            tmxr_control_line (&mpt_ldsc [port],        /*   then drop DTR on the port */
                               No_Signals);
    }

return;
}


/* Control the port datasets collectively.

   This routine is called to set the Data Terminal Ready line of all devices
   connected to multiplexer ports.  The "control" parameter indicates whether
   DTR should be raised or dropped.
*/

static void control_terminal (TMCKT control)
{
uint32 port;

tprintf (mpt_dev, TRACE_CSRW, "DTR %s on all ports\n",
         (control == DTR_Control ? "asserted" : "denied"));

for (port = 0; port < MPT_PORTS; port++)                /* loop through the set of ports */
    tmxr_control_line (&mpt_ldsc [port], control);      /*   and raise or drop DTR as directed */

return;
}
